import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import Normal

LOG_SIG_MAX = 2
LOG_SIG_MIN = -20
epsilon = 1e-6


def weights_init_(m):
    if isinstance(m, nn.Linear):
        nn.init.xavier_uniform_(m.weight, gain=1)
        nn.init.constant_(m.bias, 0)


class DDPGActor(nn.Module):
    def __init__(self, obs_size, act_size):
        super(DDPGActor, self).__init__()
        self.net = nn.Sequential(
            nn.Linear(obs_size, 400),
            nn.ReLU(),
            nn.Linear(400, 300),
            nn.ReLU(),
            nn.Linear(300, act_size),
            nn.Tanh()
        )

    def forward(self, x):
        return self.net(x)

    def get_action(self, state):
        x = torch.FloatTensor(state)
        action = self.forward(x)
        action = action.cpu().detach().numpy()
        return action


class GaussianPolicy(nn.Module):
    def __init__(self, num_inputs, num_actions, hidden_dim=256, action_space=None):
        super(GaussianPolicy, self).__init__()

        self.linear1 = nn.Linear(num_inputs, hidden_dim)
        self.linear2 = nn.Linear(hidden_dim, hidden_dim)

        self.mean_linear = nn.Linear(hidden_dim, num_actions)
        self.log_std_linear = nn.Linear(hidden_dim, num_actions)

        self.apply(weights_init_)

        # action rescaling
        if action_space is None:
            self.action_scale = torch.tensor(1.)
            self.action_bias = torch.tensor(0.)
        else:
            self.action_scale = torch.FloatTensor(
                (action_space.high - action_space.low) / 2.)
            self.action_bias = torch.FloatTensor(
                (action_space.high + action_space.low) / 2.)

    def _forward(self, state):
        x = F.relu(self.linear1(state))
        x = F.relu(self.linear2(x))
        mean = self.mean_linear(x)
        log_std = self.log_std_linear(x)
        log_std = torch.clamp(log_std, min=LOG_SIG_MIN, max=LOG_SIG_MAX)
        return mean, log_std

    def forward(self, state):
        mean, log_std = self._forward(state)
        std = log_std.exp()
        normal = Normal(mean, std)
        x_t = normal.rsample()  # for reparameterization trick (mean + std * N(0,1))
        y_t = torch.tanh(x_t)
        action = y_t * self.action_scale + self.action_bias
        return action

    def to(self, device):
        self.action_scale = self.action_scale.to(device)
        self.action_bias = self.action_bias.to(device)
        return super(GaussianPolicy, self).to(device)



import numpy as np
from typing import Dict
from rsoccer_gym.Entities.Ball import Ball
from rsoccer_gym.Entities.Robot import Robot


class Frame:
    """Units: seconds, m, m/s, degrees, degrees/s. Reference is field center."""

    def __init__(self):
        """Init Frame object."""
        self.ball: Ball = Ball()
        self.robots_blue: Dict[int, Robot] = {}
        self.robots_yellow: Dict[int, Robot] = {}


class FrameVSS(Frame):
    def parse(self, state, n_blues=3, n_yellows=3):
        """It parses the state received from grSim in a common state for environment"""
        self.ball.x = state[0]
        self.ball.y = state[1]
        self.ball.z = state[2]
        self.ball.v_x = state[3]
        self.ball.v_y = state[4]

        rbt_obs = 6
        
        for i in range(n_blues):
            robot = Robot()
            robot.id = i
            robot.x = state[5 + (rbt_obs*i) + 0]
            robot.y = state[5 + (rbt_obs*i) + 1]
            robot.theta = state[5 + (rbt_obs*i) + 2]
            robot.v_x = state[5 + (rbt_obs*i) + 3]
            robot.v_y = state[5 + (rbt_obs*i) + 4]
            robot.v_theta = state[5 + (rbt_obs*i) + 5]
            self.robots_blue[robot.id] = robot

        for i in range(n_yellows):
            robot = Robot()
            robot.id = i
            robot.x = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 0]
            robot.y = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 1]
            robot.theta = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 2]
            robot.v_x = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 3]
            robot.v_y = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 4]
            robot.v_theta = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 5]
            
            self.robots_yellow[robot.id] = robot


class FrameSSL(Frame):
    def parse(self, state, n_blues=3, n_yellows=3):
        """It parses the state received from grSim in a common state for environment"""
        self.ball.x = state[0]
        self.ball.y = state[1]
        self.ball.z = state[2]
        self.ball.v_x = state[3]
        self.ball.v_y = state[4]

        rbt_obs = 11
        
        for i in range(n_blues):
            robot = Robot()
            robot.id = i
            robot.x = state[5 + (rbt_obs*i) + 0]
            robot.y = state[5 + (rbt_obs*i) + 1]
            robot.theta = state[5 + (rbt_obs*i) + 2]
            robot.v_x = state[5 + (rbt_obs*i) + 3]
            robot.v_y = state[5 + (rbt_obs*i) + 4]
            robot.v_theta = state[5 + (rbt_obs*i) + 5]
            robot.infrared = bool(state[5 + (rbt_obs*i) + 6])
            robot.v_wheel0 = state[5 + (rbt_obs*i) + 7]
            robot.v_wheel1 = state[5 + (rbt_obs*i) + 8]
            robot.v_wheel2 = state[5 + (rbt_obs*i) + 9]
            robot.v_wheel3 = state[5 + (rbt_obs*i) + 10]
            self.robots_blue[robot.id] = robot

        for i in range(n_yellows):
            robot = Robot()
            robot.id = i
            robot.x = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 0]
            robot.y = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 1]
            robot.theta = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 2]
            robot.v_x = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 3]
            robot.v_y = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 4]
            robot.v_theta = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 5]
            robot.infrared = bool(state[5 + n_blues*rbt_obs + (rbt_obs*i) + 6])
            robot.v_wheel0 = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 7]
            robot.v_wheel1 = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 8]
            robot.v_wheel2 = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 9]
            robot.v_wheel3 = state[5 + n_blues*rbt_obs + (rbt_obs*i) + 10]
            self.robots_yellow[robot.id] = robot


class FramePB(Frame):
    def parse(self, packet):
        """It parses the state received from grSim in a common state for environment"""

        self.ball.x = packet.frame.ball.x
        self.ball.y = packet.frame.ball.y
        self.ball.v_x = packet.frame.ball.vx
        self.ball.v_y = packet.frame.ball.vy

        for _robot in packet.frame.robots_blue:
            robot = Robot()
            robot.id = _robot.robot_id
            robot.x = _robot.x
            robot.y = _robot.y
            robot.theta = np.rad2deg(_robot.orientation)
            robot.v_x = _robot.vx
            robot.v_y = _robot.vy
            robot.v_theta = np.rad2deg(_robot.vorientation)

            self.robots_blue[robot.id] = robot

        for _robot in packet.frame.robots_yellow:
            robot = Robot()
            robot.id = _robot.robot_id
            robot.x = _robot.x
            robot.y = _robot.y
            robot.theta = np.rad2deg(_robot.orientation)
            robot.v_x = _robot.vx
            robot.v_y = _robot.vy
            robot.v_theta = np.rad2deg(_robot.vorientation)

            self.robots_yellow[robot.id] = robot



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: state.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='state.proto',
  package='vss_state',
  serialized_pb=_b('\n\x0bstate.proto\x12\tvss_state\"&\n\x03RGB\x12\t\n\x01r\x18\x01 \x02(\r\x12\t\n\x01g\x18\x02 \x02(\r\x12\t\n\x01\x62\x18\x03 \x02(\r\")\n\x04Pose\x12\t\n\x01x\x18\x01 \x02(\x02\x12\t\n\x01y\x18\x02 \x02(\x02\x12\x0b\n\x03yaw\x18\x03 \x01(\x02\"\x90\x01\n\nBall_State\x12\x1d\n\x04pose\x18\x01 \x02(\x0b\x32\x0f.vss_state.Pose\x12\x1f\n\x06v_pose\x18\x02 \x01(\x0b\x32\x0f.vss_state.Pose\x12\x1f\n\x06k_pose\x18\x03 \x01(\x0b\x32\x0f.vss_state.Pose\x12!\n\x08k_v_pose\x18\x04 \x01(\x0b\x32\x0f.vss_state.Pose\"\xb0\x01\n\x0bRobot_State\x12\x1d\n\x04pose\x18\x01 \x02(\x0b\x32\x0f.vss_state.Pose\x12\x1f\n\x06v_pose\x18\x02 \x01(\x0b\x32\x0f.vss_state.Pose\x12\x1f\n\x06k_pose\x18\x03 \x01(\x0b\x32\x0f.vss_state.Pose\x12!\n\x08k_v_pose\x18\x04 \x01(\x0b\x32\x0f.vss_state.Pose\x12\x1d\n\x05\x63olor\x18\x05 \x01(\x0b\x32\x0e.vss_state.RGB\"\x9f\x02\n\x0cGlobal_State\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0e\n\x06origin\x18\x02 \x01(\x08\x12$\n\x05\x62\x61lls\x18\x03 \x03(\x0b\x32\x15.vss_state.Ball_State\x12-\n\rrobots_yellow\x18\x04 \x03(\x0b\x32\x16.vss_state.Robot_State\x12+\n\x0brobots_blue\x18\x05 \x03(\x0b\x32\x16.vss_state.Robot_State\x12\x11\n\tsituation\x18\x06 \x01(\r\x12\x0c\n\x04time\x18\x07 \x01(\r\x12\x14\n\x0cgoals_yellow\x18\x08 \x01(\r\x12\x12\n\ngoals_blue\x18\t \x01(\r\x12\x13\n\x0bname_yellow\x18\n \x01(\t\x12\x11\n\tname_blue\x18\x0b \x01(\t')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_RGB = _descriptor.Descriptor(
  name='RGB',
  full_name='vss_state.RGB',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='r', full_name='vss_state.RGB.r', index=0,
      number=1, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='g', full_name='vss_state.RGB.g', index=1,
      number=2, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='b', full_name='vss_state.RGB.b', index=2,
      number=3, type=13, cpp_type=3, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=26,
  serialized_end=64,
)


_POSE = _descriptor.Descriptor(
  name='Pose',
  full_name='vss_state.Pose',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='vss_state.Pose.x', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='y', full_name='vss_state.Pose.y', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='yaw', full_name='vss_state.Pose.yaw', index=2,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=66,
  serialized_end=107,
)


_BALL_STATE = _descriptor.Descriptor(
  name='Ball_State',
  full_name='vss_state.Ball_State',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pose', full_name='vss_state.Ball_State.pose', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='v_pose', full_name='vss_state.Ball_State.v_pose', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='k_pose', full_name='vss_state.Ball_State.k_pose', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='k_v_pose', full_name='vss_state.Ball_State.k_v_pose', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=110,
  serialized_end=254,
)


_ROBOT_STATE = _descriptor.Descriptor(
  name='Robot_State',
  full_name='vss_state.Robot_State',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pose', full_name='vss_state.Robot_State.pose', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='v_pose', full_name='vss_state.Robot_State.v_pose', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='k_pose', full_name='vss_state.Robot_State.k_pose', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='k_v_pose', full_name='vss_state.Robot_State.k_v_pose', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='color', full_name='vss_state.Robot_State.color', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=257,
  serialized_end=433,
)


_GLOBAL_STATE = _descriptor.Descriptor(
  name='Global_State',
  full_name='vss_state.Global_State',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='vss_state.Global_State.id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='origin', full_name='vss_state.Global_State.origin', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='balls', full_name='vss_state.Global_State.balls', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='robots_yellow', full_name='vss_state.Global_State.robots_yellow', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='robots_blue', full_name='vss_state.Global_State.robots_blue', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='situation', full_name='vss_state.Global_State.situation', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time', full_name='vss_state.Global_State.time', index=6,
      number=7, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='goals_yellow', full_name='vss_state.Global_State.goals_yellow', index=7,
      number=8, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='goals_blue', full_name='vss_state.Global_State.goals_blue', index=8,
      number=9, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name_yellow', full_name='vss_state.Global_State.name_yellow', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name_blue', full_name='vss_state.Global_State.name_blue', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=436,
  serialized_end=723,
)

_BALL_STATE.fields_by_name['pose'].message_type = _POSE
_BALL_STATE.fields_by_name['v_pose'].message_type = _POSE
_BALL_STATE.fields_by_name['k_pose'].message_type = _POSE
_BALL_STATE.fields_by_name['k_v_pose'].message_type = _POSE
_ROBOT_STATE.fields_by_name['pose'].message_type = _POSE
_ROBOT_STATE.fields_by_name['v_pose'].message_type = _POSE
_ROBOT_STATE.fields_by_name['k_pose'].message_type = _POSE
_ROBOT_STATE.fields_by_name['k_v_pose'].message_type = _POSE
_ROBOT_STATE.fields_by_name['color'].message_type = _RGB
_GLOBAL_STATE.fields_by_name['balls'].message_type = _BALL_STATE
_GLOBAL_STATE.fields_by_name['robots_yellow'].message_type = _ROBOT_STATE
_GLOBAL_STATE.fields_by_name['robots_blue'].message_type = _ROBOT_STATE
DESCRIPTOR.message_types_by_name['RGB'] = _RGB
DESCRIPTOR.message_types_by_name['Pose'] = _POSE
DESCRIPTOR.message_types_by_name['Ball_State'] = _BALL_STATE
DESCRIPTOR.message_types_by_name['Robot_State'] = _ROBOT_STATE
DESCRIPTOR.message_types_by_name['Global_State'] = _GLOBAL_STATE

RGB = _reflection.GeneratedProtocolMessageType('RGB', (_message.Message,), dict(
  DESCRIPTOR = _RGB,
  __module__ = 'state_pb2'
  # @@protoc_insertion_point(class_scope:vss_state.RGB)
  ))
_sym_db.RegisterMessage(RGB)

Pose = _reflection.GeneratedProtocolMessageType('Pose', (_message.Message,), dict(
  DESCRIPTOR = _POSE,
  __module__ = 'state_pb2'
  # @@protoc_insertion_point(class_scope:vss_state.Pose)
  ))
_sym_db.RegisterMessage(Pose)

Ball_State = _reflection.GeneratedProtocolMessageType('Ball_State', (_message.Message,), dict(
  DESCRIPTOR = _BALL_STATE,
  __module__ = 'state_pb2'
  # @@protoc_insertion_point(class_scope:vss_state.Ball_State)
  ))
_sym_db.RegisterMessage(Ball_State)

Robot_State = _reflection.GeneratedProtocolMessageType('Robot_State', (_message.Message,), dict(
  DESCRIPTOR = _ROBOT_STATE,
  __module__ = 'state_pb2'
  # @@protoc_insertion_point(class_scope:vss_state.Robot_State)
  ))
_sym_db.RegisterMessage(Robot_State)

Global_State = _reflection.GeneratedProtocolMessageType('Global_State', (_message.Message,), dict(
  DESCRIPTOR = _GLOBAL_STATE,
  __module__ = 'state_pb2'
  # @@protoc_insertion_point(class_scope:vss_state.Global_State)
  ))
_sym_db.RegisterMessage(Global_State)


# @@protoc_insertion_point(module_scope)



import torch
import torch.nn as nn


class DDPGActor(nn.Module):
    def __init__(self, obs_size, act_size):
        super(DDPGActor, self).__init__()
        self.net = nn.Sequential(
            nn.Linear(obs_size, 400),
            nn.ReLU(),
            nn.Linear(400, 300),
            nn.ReLU(),
            nn.Linear(300, act_size),
            nn.Tanh()
        )

    def forward(self, x):
        return self.net(x)

    def get_action(self, state):
        x = torch.FloatTensor(state)
        action = self.forward(x)
        action = action.cpu().detach().numpy()
        return action



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: common.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='common.proto',
  package='fira_message',
  syntax='proto3',
  serialized_pb=_b('\n\x0c\x63ommon.proto\x12\x0c\x66ira_message\"K\n\x04\x42\x61ll\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\x12\t\n\x01z\x18\x03 \x01(\x01\x12\n\n\x02vx\x18\x04 \x01(\x01\x12\n\n\x02vy\x18\x05 \x01(\x01\x12\n\n\x02vz\x18\x06 \x01(\x01\"r\n\x05Robot\x12\x10\n\x08robot_id\x18\x01 \x01(\r\x12\t\n\x01x\x18\x02 \x01(\x01\x12\t\n\x01y\x18\x03 \x01(\x01\x12\x13\n\x0borientation\x18\x04 \x01(\x01\x12\n\n\x02vx\x18\x05 \x01(\x01\x12\n\n\x02vy\x18\x06 \x01(\x01\x12\x14\n\x0cvorientation\x18\x07 \x01(\x01\"N\n\x05\x46ield\x12\r\n\x05width\x18\x01 \x01(\x01\x12\x0e\n\x06length\x18\x02 \x01(\x01\x12\x12\n\ngoal_width\x18\x03 \x01(\x01\x12\x12\n\ngoal_depth\x18\x04 \x01(\x01\"\x7f\n\x05\x46rame\x12 \n\x04\x62\x61ll\x18\x01 \x01(\x0b\x32\x12.fira_message.Ball\x12*\n\rrobots_yellow\x18\x02 \x03(\x0b\x32\x13.fira_message.Robot\x12(\n\x0brobots_blue\x18\x03 \x03(\x0b\x32\x13.fira_message.Robotb\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_BALL = _descriptor.Descriptor(
  name='Ball',
  full_name='fira_message.Ball',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='fira_message.Ball.x', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='y', full_name='fira_message.Ball.y', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='z', full_name='fira_message.Ball.z', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='vx', full_name='fira_message.Ball.vx', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='vy', full_name='fira_message.Ball.vy', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='vz', full_name='fira_message.Ball.vz', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=30,
  serialized_end=105,
)


_ROBOT = _descriptor.Descriptor(
  name='Robot',
  full_name='fira_message.Robot',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='robot_id', full_name='fira_message.Robot.robot_id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='x', full_name='fira_message.Robot.x', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='y', full_name='fira_message.Robot.y', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='orientation', full_name='fira_message.Robot.orientation', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='vx', full_name='fira_message.Robot.vx', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='vy', full_name='fira_message.Robot.vy', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='vorientation', full_name='fira_message.Robot.vorientation', index=6,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=107,
  serialized_end=221,
)


_FIELD = _descriptor.Descriptor(
  name='Field',
  full_name='fira_message.Field',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='width', full_name='fira_message.Field.width', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='length', full_name='fira_message.Field.length', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='goal_width', full_name='fira_message.Field.goal_width', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='goal_depth', full_name='fira_message.Field.goal_depth', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=223,
  serialized_end=301,
)


_FRAME = _descriptor.Descriptor(
  name='Frame',
  full_name='fira_message.Frame',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ball', full_name='fira_message.Frame.ball', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='robots_yellow', full_name='fira_message.Frame.robots_yellow', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='robots_blue', full_name='fira_message.Frame.robots_blue', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=303,
  serialized_end=430,
)

_FRAME.fields_by_name['ball'].message_type = _BALL
_FRAME.fields_by_name['robots_yellow'].message_type = _ROBOT
_FRAME.fields_by_name['robots_blue'].message_type = _ROBOT
DESCRIPTOR.message_types_by_name['Ball'] = _BALL
DESCRIPTOR.message_types_by_name['Robot'] = _ROBOT
DESCRIPTOR.message_types_by_name['Field'] = _FIELD
DESCRIPTOR.message_types_by_name['Frame'] = _FRAME

Ball = _reflection.GeneratedProtocolMessageType('Ball', (_message.Message,), dict(
  DESCRIPTOR = _BALL,
  __module__ = 'common_pb2'
  # @@protoc_insertion_point(class_scope:fira_message.Ball)
  ))
_sym_db.RegisterMessage(Ball)

Robot = _reflection.GeneratedProtocolMessageType('Robot', (_message.Message,), dict(
  DESCRIPTOR = _ROBOT,
  __module__ = 'common_pb2'
  # @@protoc_insertion_point(class_scope:fira_message.Robot)
  ))
_sym_db.RegisterMessage(Robot)

Field = _reflection.GeneratedProtocolMessageType('Field', (_message.Message,), dict(
  DESCRIPTOR = _FIELD,
  __module__ = 'common_pb2'
  # @@protoc_insertion_point(class_scope:fira_message.Field)
  ))
_sym_db.RegisterMessage(Field)

Frame = _reflection.GeneratedProtocolMessageType('Frame', (_message.Message,), dict(
  DESCRIPTOR = _FRAME,
  __module__ = 'common_pb2'
  # @@protoc_insertion_point(class_scope:fira_message.Frame)
  ))
_sym_db.RegisterMessage(Frame)


# @@protoc_insertion_point(module_scope)



'''
#   Environment Communication Structure
#    - Father class that creates the structure to communicate with multples setups of enviroment
#    - To create your wrapper from env to communcation, use inherit from this class! 
'''


import time
from typing import Dict, List, Optional

import gym
import numpy as np
from rsoccer_gym.Entities import Frame, Robot
from rsoccer_gym.Simulators.rsim import RSimVSS
from rsoccer_gym.Simulators.fira import Fira



class VSSBaseEnv(gym.Env):
    metadata = {
        'render.modes': ['human', 'rgb_array'],
    }
    NORM_BOUNDS = 1.2
    
    def __init__(self, field_type: int,
                 n_robots_blue: int, n_robots_yellow: int, time_step: float):
        # Initialize Simulator
        self.time_step = time_step
        self.rsim = RSimVSS(field_type=field_type,
                                      n_robots_blue=n_robots_blue,
                                      n_robots_yellow=n_robots_yellow,
                                      time_step_ms=int(self.time_step*1000))
        self.n_robots_blue = n_robots_blue
        self.n_robots_yellow = n_robots_yellow

        # Get field dimensions
        self.field_type = field_type
        self.field = self.rsim.get_field_params()
        self.max_pos = max(self.field.width / 2, (self.field.length / 2) 
                                + self.field.penalty_length)
        max_wheel_rad_s = (self.field.rbt_motor_max_rpm / 60) * 2 * np.pi
        self.max_v = max_wheel_rad_s * self.field.rbt_wheel_radius
        # 0.04 = robot radius (0.0375) + wheel thicknees (0.0025)
        self.max_w = np.rad2deg(self.max_v / 0.04)
        
        # Initiate 
        self.frame: Frame = None
        self.last_frame: Frame = None
        self.view = None
        self.steps = 0
        self.sent_commands = None

    def step(self, action):
        self.steps += 1
        # Join agent action with environment actions
        commands: List[Robot] = self._get_commands(action)
        # Send command to simulator
        self.rsim.send_commands(commands)
        self.sent_commands = commands

        # Get Frame from simulator
        self.last_frame = self.frame
        self.frame = self.rsim.get_frame()

        # Calculate environment observation, reward and done condition
        observation = self._frame_to_observations()
        reward, done = self._calculate_reward_and_done()

        return observation, reward, done, {}

    def reset(self):
        self.steps = 0
        self.last_frame = None
        self.sent_commands = None

        # Close render window
        del(self.view)
        self.view = None

        initial_pos_frame: Frame = self._get_initial_positions_frame()
        self.rsim.reset(initial_pos_frame)

        # Get frame from simulator
        self.frame = self.rsim.get_frame()

        return self._frame_to_observations()

    def render(self, mode='human') -> None:
        '''
        Renders the game depending on 
        ball's and players' positions.

        Parameters
        ----------
        None

        Returns
        -------
        None

        '''
        if self.view == None:
            from rsoccer_gym.Render import RCGymRender
            self.view = RCGymRender(self.n_robots_blue,
                                 self.n_robots_yellow,
                                 self.field,
                                 simulator='vss')

        return self.view.render_frame(self.frame, return_rgb_array=mode == "rgb_array")
        

    def close(self):
        self.rsim.stop()

    def _get_commands(self, action):
        '''returns a list of commands of type List[Robot] from type action_space action'''
        raise NotImplementedError

    def _frame_to_observations(self):
        '''returns a type observation_space observation from a type List[Robot] state'''
        raise NotImplementedError

    def _calculate_reward_and_done(self):
        '''returns reward value and done flag from type List[Robot] state'''
        raise NotImplementedError

    def _get_initial_positions_frame(self) -> Frame:
        '''returns frame with robots initial positions'''
        raise NotImplementedError

    def norm_pos(self, pos):
        return np.clip(
            pos / self.max_pos,
            -self.NORM_BOUNDS,
            self.NORM_BOUNDS
        )

    def norm_v(self, v):
        return np.clip(
            v / self.max_v,
            -self.NORM_BOUNDS,
            self.NORM_BOUNDS
        )

    def norm_w(self, w):
        return np.clip(
            w / self.max_w,
            -self.NORM_BOUNDS,
            self.NORM_BOUNDS
        )


class VSSBaseFIRAEnv(VSSBaseEnv):

    def __init__(self, field_type: int,
                 n_robots_blue: int, n_robots_yellow: int, time_step: float):
        super().__init__(field_type, n_robots_blue, n_robots_yellow, time_step)
        self.rsim = Fira()



from dataclasses import dataclass
import numpy as np

@dataclass()
class Robot:
    yellow: bool = None
    id: int = None
    x: float = None
    y: float = None
    z: float = None
    theta: float = None
    v_x: float = 0
    v_y: float = 0
    v_theta: float = 0
    kick_v_x: float = 0
    kick_v_z: float = 0
    dribbler: bool = False
    infrared: bool = False
    wheel_speed: bool = False
    v_wheel0: float = 0 # rad/s
    v_wheel1: float = 0 # rad/s
    v_wheel2: float = 0 # rad/s
    v_wheel3: float = 0 # rad/s


import math
import random
from typing import Dict

import gym
import numpy as np
from rsoccer_gym.Entities import Ball, Frame, Robot
from rsoccer_gym.ssl.ssl_gym_base import SSLBaseEnv


class SSLPassEnduranceMAEnv(SSLBaseEnv):
    """The SSL robot needs to make a goal with contested possession


        Description:

        Observation:
            Type: Box(4 + 6*n_robots_blue)
            Normalized Bounds to [-1.2, 1.2]
            Num      Observation normalized  
            0->3     Ball [X, Y, V_x, V_y]
            4->13    id N Blue [X, Y, sin(theta), cos(theta), V_theta]


        Actions:
            Type: Box(3,)
            Num     Action
            0       id i Blue Angular Speed  (%)
            1       id i Blue Kick x Speed  (%)
            2       id i Blue Dribbler  (%) (true if % is positive)

        Reward:

        Starting State:

        Episode Termination:
            30 seconds (1200 steps) or wrong pass
    """

    def __init__(self):
        super().__init__(field_type=2, n_robots_blue=2,
                         n_robots_yellow=0, time_step=0.025)
        self.action_space = gym.spaces.Box(low=-1, high=1,
                                           shape=(self.n_robots_blue, 5),
                                           dtype=np.float32)

        n_obs = 4 + 9*self.n_robots_blue
        self.stopped_steps = 0
        self.observation_space = gym.spaces.Box(low=-self.NORM_BOUNDS,
                                                high=self.NORM_BOUNDS,
                                                shape=(
                                                    self.n_robots_blue, n_obs),
                                                dtype=np.float32)
        self.receiver_id = 1
        self.shooter_id = 0
        self.ball_grad_scale = np.linalg.norm([self.field.width/2,
                                               self.field.length/2])/4
        wheel_max_rad_s = 160
        max_steps = 1200
        self.energy_scale = ((wheel_max_rad_s * 4) * max_steps)

        # Limit robot speeds
        self.max_v = 2.5
        self.max_w = 10
        self.max_kick_x = 5.0

        print('Environment initialized')

    def get_rotated_obs(self):
        robots_dict = dict()
        for i in range(self.n_robots_blue):
            robots_dict[i] = list()
            robots_dict[i].append(self.norm_pos(self.frame.robots_blue[i].x))
            robots_dict[i].append(self.norm_pos(self.frame.robots_blue[i].y))
            robots_dict[i].append(
                np.sin(np.deg2rad(self.frame.robots_blue[i].theta))
            )
            robots_dict[i].append(
                np.cos(np.deg2rad(self.frame.robots_blue[i].theta))
            )
            robots_dict[i].append(self.norm_v(self.frame.robots_blue[i].v_x))
            robots_dict[i].append(self.norm_v(self.frame.robots_blue[i].v_y))
            robots_dict[i].append(self.norm_w(
                self.frame.robots_blue[i].v_theta))
            robots_dict[i].append(
                1 if self.frame.robots_blue[i].infrared else 0)
            robots_dict[i].append(1 if i == self.shooter_id else 0)

        rotaded_obs = list()
        for i in range(self.n_robots_blue):
            aux_dict = {}
            aux_dict.update(robots_dict)
            rotated = list()
            rotated = rotated + aux_dict.pop(i)
            teammates = list(aux_dict.values())
            for teammate in teammates:
                rotated = rotated + teammate
            rotaded_obs.append(rotated)

        return rotaded_obs

    def _frame_to_observations(self):

        observations = list()
        robots = self.get_rotated_obs()
        for idx in range(self.n_robots_blue):
            observation = []
            observation.append(self.norm_pos(self.frame.ball.x))
            observation.append(self.norm_pos(self.frame.ball.y))
            observation.append(self.norm_v(self.frame.ball.v_x))
            observation.append(self.norm_v(self.frame.ball.v_y))
            observation += robots[idx]
            observations.append(np.array(observation, dtype=np.float32))

        observations = np.array(observations)
        return observations

    def reset(self):
        self.reward_shaping_total = None
        state = super().reset()
        self.stopped_steps = 0
        self.shooter_id, self.receiver_id = 0, 1
        return state

    def step(self, action):
        observation, reward, done, _ = super().step(action)
        return observation, reward, done, self.reward_shaping_total

    def _get_commands(self, actions):
        commands = []
        actions[0][3] = actions[0][3] if abs(actions[0][3]) > 0.5 else 0
        actions[1][3] = actions[1][3] if abs(actions[1][3]) > 0.5 else 0


        for i in range(self.n_robots_blue):
            angle = self.frame.robots_blue[i].theta
            v_x, v_y, v_theta = self.convert_actions(actions[i], np.deg2rad(angle))
            cmd = Robot(yellow=False, id=i, v_x=v_x, v_y=v_y, v_theta=v_theta,
                        kick_v_x=actions[i][3] * self.max_kick_x,
                        dribbler=True if actions[i][4] > 0 else False)
            commands.append(cmd)

        return commands
    
    def convert_actions(self, action, angle):
        """Denormalize, clip to absolute max and convert to local"""
        # Denormalize
        v_x = action[0] * self.max_v
        v_y = action[1] * self.max_v
        v_theta = action[2] * self.max_w

        # Convert to local
        v_x, v_y = v_x*np.cos(angle) + v_y*np.sin(angle),\
            -v_x*np.sin(angle) + v_y*np.cos(angle)

        # clip by max absolute
        v_norm = np.linalg.norm([v_x,v_y])
        c = v_norm < self.max_v or self.max_v / v_norm
        v_x, v_y = v_x*c, v_y*c
        
        return v_x, v_y, v_theta

    def _calculate_reward_and_done(self):
        w_ball_grad = 1/self.ball_grad_scale
        w_energy = 1/self.energy_scale
        reward = {f'robot_{i}': 0 for i in range(self.n_robots_blue)}
        done = False
        if self.reward_shaping_total is None:
            self.reward_shaping_total = {'n_passes': 0,
                                         'ball_grad': 0}
            for i in range(self.n_robots_blue):
                self.reward_shaping_total[f'robot_{i}'] = {'energy': 0}

        if self.frame.robots_blue[self.receiver_id].infrared:
            for i in range(self.n_robots_blue):
                reward[f'robot_{i}'] = 10
            self.reward_shaping_total['n_passes'] += 1
            self.stopped_steps = 0
            self.shooter_id, self.receiver_id = self.receiver_id, self.shooter_id
        else:
            rw_ball_grad = w_ball_grad * self.__ball_grad_rw()
            reward[f'robot_{self.shooter_id}'] += rw_ball_grad
            reward[f'robot_{self.receiver_id}'] += rw_ball_grad
            self.reward_shaping_total['ball_grad'] += rw_ball_grad

            for i in range(self.n_robots_blue):
                rw_energy = w_energy*self.__energy_pen(i)
                reward[f'robot_{i}'] += rw_energy
                self.reward_shaping_total[f'robot_{i}']['energy'] += rw_energy

        if self.__bad_state():
            for i in range(self.n_robots_blue):
                reward[f'robot_{i}'] = -1
            done = True
        return reward, done

    def _get_initial_positions_frame(self):
        '''Returns the position of each robot and ball for the initial frame'''
        pos_frame: Frame = Frame()
        def x(): return random.uniform(-2, 2)
        def y(): return random.uniform(1.5, -1.5)

        pos_frame.ball = Ball(x=x(), y=y())
        factor = (pos_frame.ball.y/abs(pos_frame.ball.y))
        offset = 0.09*factor
        angle = 270 if factor > 0 else 90
        pos_frame.robots_blue[0] = Robot(
            x=pos_frame.ball.x, y=pos_frame.ball.y+offset, theta=angle
        )
        ball = np.array([pos_frame.ball.x,
                         pos_frame.ball.y])
        recv_x = x()
        while abs(recv_x - pos_frame.ball.x) < 1.5:
            recv_x = x()
        receiver = np.array([recv_x, -pos_frame.ball.y])
        vect = receiver - ball
        recv_angle = np.rad2deg(np.arctan2(vect[1], vect[0]) + np.pi)

        pos_frame.robots_blue[1] = Robot(x=receiver[0],
                                         y=receiver[1],
                                         theta=recv_angle)

        return pos_frame

    def __bad_state(self):
        # Check if dist between robots > 1.5
        recv = np.array([self.frame.robots_blue[self.receiver_id].x,
                         self.frame.robots_blue[self.receiver_id].y])
        shooter = np.array([self.frame.robots_blue[self.shooter_id].x,
                            self.frame.robots_blue[self.shooter_id].y])
        min_dist = np.linalg.norm(recv - shooter) > 1.5

        # Check if ball is in this rectangle
        ball = np.array([self.frame.ball.x, self.frame.ball.y])
        last_ball = np.array([self.last_frame.ball.x, self.last_frame.ball.y])
        inside = -2 < ball[0] < 2 and -1.5 < ball[1] < 1.5

        # Check if ball is stopped for too long
        last_dist = np.linalg.norm(last_ball - recv)
        dist = np.linalg.norm(ball - recv)
        stopped = abs(last_dist - dist) < 0.01
        if stopped:
            self.stopped_steps += 1
        else:
            self.stopped_steps = 0
        return self.stopped_steps > 20 or not inside or not min_dist

    def __ball_grad_rw(self):
        assert(self.last_frame is not None)

        # Goal pos
        goal = np.array([self.frame.robots_blue[self.receiver_id].x,
                         self.frame.robots_blue[self.receiver_id].y])

        # Calculate previous ball dist
        last_ball = self.last_frame.ball
        ball = self.frame.ball
        last_ball_pos = np.array([last_ball.x, last_ball.y])
        last_ball_dist = np.linalg.norm(goal - last_ball_pos)

        # Calculate new ball dist
        ball_pos = np.array([ball.x, ball.y])
        ball_dist = np.linalg.norm(goal - ball_pos)

        ball_dist_rw = last_ball_dist - ball_dist
        return np.clip(ball_dist_rw, -1, 1)

    def __energy_pen(self, idx):
        robot = self.frame.robots_blue[idx]

        # Sum of abs each wheel speed sent
        energy = abs(robot.v_wheel0)\
            + abs(robot.v_wheel1)\
            + abs(robot.v_wheel2)\
            + abs(robot.v_wheel3)

        return energy



import numpy as np
import robosim

from typing import Dict, List
from rsoccer_gym.Entities import Frame, FrameVSS, FrameSSL, Field


class RSim:
    def __init__(
        self,
        field_type: int,
        n_robots_blue: int,
        n_robots_yellow: int,
        time_step_ms: int,
    ):
        self.n_robots_blue = n_robots_blue
        self.n_robots_yellow = n_robots_yellow

        # Positions needed just to initialize the simulator
        ball_pos = [0, 0, 0, 0]
        blue_robots_pos = [[-0.2 * i, 0, 0]
                           for i in range(1, n_robots_blue + 1)]
        yellow_robots_pos = [[0.2 * i, 0, 0]
                             for i in range(1, n_robots_yellow + 1)]
        self.simulator = self._init_simulator(
            field_type=field_type,
            n_robots_blue=n_robots_blue,
            n_robots_yellow=n_robots_yellow,
            ball_pos=ball_pos,
            blue_robots_pos=blue_robots_pos,
            yellow_robots_pos=yellow_robots_pos,
            time_step_ms=time_step_ms
        )
        self.field = self.get_field_params()

    def reset(self, frame: Frame):
        placement_pos = self._placement_dict_from_frame(frame)
        self.simulator.reset(placement_pos["ball_pos"], placement_pos["blue_robots_pos"], placement_pos["yellow_robots_pos"])

    def stop(self):
        del self.simulator

    def send_commands(self, commands):
        raise NotImplementedError

    def get_frame(self) -> Frame:
        raise NotImplementedError

    def get_field_params(self):
        return Field(**self.simulator.get_field_params())
    
    def _placement_dict_from_frame(self, frame: Frame):
        replacement_pos: Dict[str, np.ndarray] = {}

        ball_pos: List[float] = [
            frame.ball.x,
            frame.ball.y,
            frame.ball.v_x,
            frame.ball.v_y,
        ]
        replacement_pos["ball_pos"] = np.array(ball_pos)

        blue_pos: List[List[float]] = []
        for robot in frame.robots_blue.values():
            robot_pos: List[float] = [robot.x, robot.y, robot.theta]
            blue_pos.append(robot_pos)
        replacement_pos["blue_robots_pos"] = np.array(blue_pos)

        yellow_pos: List[List[float]] = []
        for robot in frame.robots_yellow.values():
            robot_pos: List[float] = [robot.x, robot.y, robot.theta]
            yellow_pos.append(robot_pos)
        replacement_pos["yellow_robots_pos"] = np.array(yellow_pos)

        return replacement_pos

    def _init_simulator(
        self,
        field_type,
        n_robots_blue,
        n_robots_yellow,
        ball_pos,
        blue_robots_pos,
        yellow_robots_pos,
        time_step_ms,
    ):
        raise NotImplementedError


class RSimVSS(RSim):
    def send_commands(self, commands):
        sim_commands = np.zeros(
            (self.n_robots_blue + self.n_robots_yellow, 2), dtype=np.float64)

        for cmd in commands:
            if cmd.yellow:
                rbt_id = self.n_robots_blue + cmd.id
            else:
                rbt_id = cmd.id
            sim_commands[rbt_id][0] = cmd.v_wheel0
            sim_commands[rbt_id][1] = cmd.v_wheel1
        self.simulator.step(sim_commands)

    def get_frame(self) -> FrameVSS:
        state = self.simulator.get_state()
        # Update frame with new state
        frame = FrameVSS()
        frame.parse(state, self.n_robots_blue, self.n_robots_yellow)

        return frame

    def _init_simulator(self, field_type, n_robots_blue, n_robots_yellow,
                        ball_pos, blue_robots_pos, yellow_robots_pos,
                        time_step_ms):

        return robosim.VSS(
            field_type,
            n_robots_blue,
            n_robots_yellow,
            time_step_ms,
            ball_pos,
            blue_robots_pos,
            yellow_robots_pos,
        )


class RSimSSL(RSim):
    def send_commands(self, commands):
        sim_cmds = np.zeros(
            (self.n_robots_blue + self.n_robots_yellow, 8), dtype=np.float64)

        for cmd in commands:
            if cmd.yellow:
                rbt_id = self.n_robots_blue + cmd.id
            else:
                rbt_id = cmd.id
            if cmd.wheel_speed:
                sim_cmds[rbt_id][0] = cmd.wheel_speed
                sim_cmds[rbt_id][1] = cmd.v_wheel0
                sim_cmds[rbt_id][2] = cmd.v_wheel1
                sim_cmds[rbt_id][3] = cmd.v_wheel2
                sim_cmds[rbt_id][4] = cmd.v_wheel3
                sim_cmds[rbt_id][5] = cmd.kick_v_x
                sim_cmds[rbt_id][6] = cmd.kick_v_z
                sim_cmds[rbt_id][7] = cmd.dribbler
            else:
                sim_cmds[rbt_id][0] = cmd.wheel_speed
                sim_cmds[rbt_id][1] = cmd.v_x
                sim_cmds[rbt_id][2] = cmd.v_y
                sim_cmds[rbt_id][3] = cmd.v_theta
                sim_cmds[rbt_id][5] = cmd.kick_v_x
                sim_cmds[rbt_id][6] = cmd.kick_v_z
                sim_cmds[rbt_id][7] = cmd.dribbler
            
        self.simulator.step(sim_cmds)

    def get_frame(self) -> FrameSSL:
        state = self.simulator.get_state()
        # Update frame with new state
        frame = FrameSSL()
        frame.parse(state, self.n_robots_blue, self.n_robots_yellow)

        return frame
    
    def _init_simulator(self, field_type, n_robots_blue, n_robots_yellow,
                        ball_pos, blue_robots_pos, yellow_robots_pos,
                        time_step_ms):

        return robosim.SSL(
            field_type,
            n_robots_blue,
            n_robots_yellow,
            time_step_ms,
            ball_pos,
            blue_robots_pos,
            yellow_robots_pos,
        )



from kdtree import KDTree
import unittest


class TestKDTree(unittest.TestCase):

    def test_kdtree(self):

        tree = KDTree()
        tree.insert((2, 6))
        tree.insert((3, 1))
        tree.insert((8, 7))
        tree.insert((10, 2))
        tree.insert((13, 3))

        assert tree.get_nearest((9, 4)) == ((10, 2), 2.23606797749979)
        assert tree.get_nearest((4, 1.5))[0] == (3, 1)
        assert tree.get_nearest((7, 8))[0] == (8, 7)
        assert tree.get_nearest((11, 1))[0] == (10, 2)
        assert tree.get_nearest((13, 3))[0] == (13, 3)


if __name__ == '__main__':
    unittest.main()



from rsoccer_gym.ssl.ssl_hw_challenge.static_defenders import SSLHWStaticDefendersEnv
from rsoccer_gym.ssl.ssl_hw_challenge.dribbling import SSLHWDribblingEnv
from rsoccer_gym.ssl.ssl_hw_challenge.contested_possession import SSLContestedPossessionEnv
from rsoccer_gym.ssl.ssl_hw_challenge.pass_endurance import SSLPassEnduranceEnv
from rsoccer_gym.ssl.ssl_hw_challenge.pass_endurance_ma import SSLPassEnduranceMAEnv



import math
import random
from typing import Dict

import gym
import numpy as np
from rsoccer_gym.Entities import Ball, Frame, Robot
from rsoccer_gym.ssl.ssl_gym_base import SSLBaseEnv


class SSLPassEnduranceEnv(SSLBaseEnv):
    """The SSL robot needs to make a goal with contested possession


        Description:

        Observation:
            Type: Box(4 + 6*n_robots_blue)
            Normalized Bounds to [-1.2, 1.2]
            Num      Observation normalized  
            0->3     Ball [X, Y, V_x, V_y]
            4->13    id N Blue [X, Y, sin(theta), cos(theta), V_theta]


        Actions:
            Type: Box(3,)
            Num     Action
            0       id i Blue Angular Speed  (%)
            1       id i Blue Kick x Speed  (%)
            2       id i Blue Dribbler  (%) (true if % is positive)

        Reward:

        Starting State:

        Episode Termination:
            30 seconds (1200 steps) or wrong pass
    """
    original_vec = None
    max_dist = 100
    actions = {}
    shooted = False

    def __init__(self):
        super().__init__(field_type=2, n_robots_blue=2,
                         n_robots_yellow=0, time_step=0.025)
        self.action_space = gym.spaces.Box(low=-1, high=1,
                                           shape=(3, ),
                                           dtype=np.float32)

        n_obs = 4 + 6*self.n_robots_blue
        self.holding_steps = 0
        self.stopped_steps = 0
        self.recv_angle = 270
        self.observation_space = gym.spaces.Box(low=-self.NORM_BOUNDS,
                                                high=self.NORM_BOUNDS,
                                                shape=(n_obs,),
                                                dtype=np.float32)
        self.receiver_id = 1
        self.ball_grad_scale = np.linalg.norm([self.field.width/2,
                                               self.field.length/2])/4

        # Limit robot speeds
        self.max_v = 2.5
        self.max_w = 10
        self.max_kick_x = 5.0

        print('Environment initialized')

    def _frame_to_observations(self):
        observation = list()

        observation.append(self.norm_pos(self.frame.ball.x))
        observation.append(self.norm_pos(self.frame.ball.y))
        observation.append(self.norm_v(self.frame.ball.v_x))
        observation.append(self.norm_v(self.frame.ball.v_y))
        for i in range(self.n_robots_blue):
            observation.append(self.norm_pos(self.frame.robots_blue[i].x))
            observation.append(self.norm_pos(self.frame.robots_blue[i].y))
            observation.append(
                np.sin(np.deg2rad(self.frame.robots_blue[i].theta))
            )
            observation.append(
                np.cos(np.deg2rad(self.frame.robots_blue[i].theta))
            )
            observation.append(self.norm_w(self.frame.robots_blue[i].v_theta))
            observation.append(1 if self.frame.robots_blue[i].infrared else 0)
        return np.array(observation, dtype=np.float32)

    def reset(self):
        self.reward_shaping_total = None
        state = super().reset()
        self.actions = {}
        self.holding_steps = 0
        self.stopped_steps = 0
        self.shooted = False
        return state

    def step(self, action):
        observation, reward, done, _ = super().step(action)
        return observation, reward, done, self.reward_shaping_total

    def _get_commands(self, actions):
        commands = []
        actions[1] = actions[1] if abs(actions[1]) > 0.5 else 0
        self.actions = actions
        cmd = Robot(yellow=False, id=0, v_x=0,
                    v_y=0, v_theta=actions[0] * self.max_w,
                    kick_v_x=actions[1] * self.max_kick_x,
                    dribbler=True if actions[2] > 0 else False)
        commands.append(cmd)

        cmd = Robot(yellow=False, id=1, v_x=0,
                    v_y=0, v_theta=0,
                    kick_v_x=0,
                    dribbler=True)
        commands.append(cmd)

        return commands

    def _calculate_reward_and_done(self):
        w_ball_grad = 1/self.ball_grad_scale
        reward = 0
        done = False
        if self.reward_shaping_total is None:
            self.reward_shaping_total = {'reversed_dist': 0,
                                         'ball_grad': 0}
        if self.frame.robots_blue[1].infrared:
            reward += 1
            done = True
        else:
            rw_ball_grad = w_ball_grad * self.__ball_grad_rw()
            reward = rw_ball_grad
            self.reward_shaping_total['ball_grad'] += rw_ball_grad
        if self.__wrong_ball() or self.holding_steps > 15:
            reward -= 1
            done = True

        if done:
            ball = np.array([self.frame.ball.x, self.frame.ball.y])
            recv = np.array([self.frame.robots_blue[1].x,
                             self.frame.robots_blue[1].y])
            shooter = np.array([self.frame.robots_blue[0].x,
                                self.frame.robots_blue[0].y])
            dist_robs = np.linalg.norm(recv - shooter)
            dist_ball = np.linalg.norm(recv - ball)
            reversed_dist = dist_robs - dist_ball
            normed = reversed_dist/dist_robs
            self.reward_shaping_total['reversed_dist'] = normed
        return reward, done

    def _get_initial_positions_frame(self):
        '''Returns the position of each robot and ball for the initial frame'''
        pos_frame: Frame = Frame()
        def x(): return random.uniform(-1.5, 1.5)
        def y(): return random.uniform(1.5, -1.5)

        pos_frame.ball = Ball(x=x(), y=y())
        factor = (pos_frame.ball.y/abs(pos_frame.ball.y))
        offset = 0.115*factor
        angle = 270 if factor > 0 else 90
        pos_frame.robots_blue[0] = Robot(
            x=pos_frame.ball.x, y=pos_frame.ball.y+offset, theta=angle
        )
        shooter = np.array([pos_frame.robots_blue[0].x,
                            pos_frame.robots_blue[0].y])
        recv_x = x()
        while abs(recv_x - pos_frame.ball.x) < 1:
            recv_x = x()
        receiver = np.array([recv_x, -pos_frame.ball.y])
        vect = receiver - shooter
        recv_angle = np.rad2deg(np.arctan2(vect[1], vect[0]) + np.pi)

        pos_frame.robots_blue[1] = Robot(x=receiver[0],
                                         y=receiver[1],
                                         theta=recv_angle)

        return pos_frame

    def __wrong_ball(self):
        ball = np.array([self.frame.ball.x, self.frame.ball.y])
        last_ball = np.array([self.last_frame.ball.x, self.last_frame.ball.y])

        recv = np.array([self.frame.robots_blue[1].x,
                         self.frame.robots_blue[1].y])
        shooter = np.array([self.frame.robots_blue[0].x,
                            self.frame.robots_blue[0].y])
        comp_ball = np.array(ball*100, dtype=int)
        comp_shoot = np.array(shooter*100, dtype=int)
        comp_recv = np.array(recv*100, dtype=int)
        inside_x = min(comp_recv[0], comp_shoot[0]) <= comp_ball[0] <= max(
            comp_recv[0], comp_shoot[0])
        inside_y = min(comp_recv[1], comp_shoot[1]) <= comp_ball[1] <= max(
            comp_recv[1], comp_shoot[1])
        not_inside = not(inside_x and inside_y)
        last_dist = np.linalg.norm(last_ball - recv)
        dist = np.linalg.norm(ball - recv)
        stopped = abs(last_dist - dist) < 0.01
        if stopped:
            self.stopped_steps += 1
        else:
            self.stopped_steps = 0
        return self.stopped_steps > 20 or not_inside

    def __ball_grad_rw(self):
        assert(self.last_frame is not None)

        # Goal pos
        goal = np.array([self.frame.robots_blue[1].x,
                         self.frame.robots_blue[1].y])

        # Calculate previous ball dist
        last_ball = self.last_frame.ball
        ball = self.frame.ball
        last_ball_pos = np.array([last_ball.x, last_ball.y])
        last_ball_dist = np.linalg.norm(goal - last_ball_pos)

        # Calculate new ball dist
        ball_pos = np.array([ball.x, ball.y])
        ball_dist = np.linalg.norm(goal - ball_pos)

        ball_dist_rw = last_ball_dist - ball_dist
        return np.clip(ball_dist_rw, -1, 1)



from gym.envs.registration import register

register(id='VSS-v0',
         entry_point='rsoccer_gym.vss.env_vss:VSSEnv',
         max_episode_steps=1200
         )

register(id='VSSMA-v0',
         entry_point='rsoccer_gym.vss.env_ma:VSSMAEnv',
         max_episode_steps=1200
         )

register(id='VSSMAOpp-v0',
         entry_point='rsoccer_gym.vss.env_ma:VSSMAOpp',
         max_episode_steps=1200
         )

register(id='VSSGk-v0',
         entry_point='rsoccer_gym.vss.env_gk:rSimVSSGK',
         max_episode_steps=1200
         )

register(id='VSSFIRA-v0',
         entry_point='rsoccer_gym.vss.env_vss:VSSFIRAEnv',
         max_episode_steps=1200
         )

register(id='SSLGoToBall-v0',
         entry_point='rsoccer_gym.ssl.ssl_go_to_ball:SSLGoToBallEnv',
         kwargs={'field_type': 2, 'n_robots_yellow': 6},
         max_episode_steps=1200
         )

register(id='SSLGoToBallIR-v0',
         entry_point='rsoccer_gym.ssl.ssl_go_to_ball:SSLGoToBallIREnv',
         kwargs={'field_type': 2, 'n_robots_yellow': 6},
         max_episode_steps=1200
         )

register(id='SSLGoToBallShoot-v0',
         entry_point='rsoccer_gym.ssl.ssl_go_to_ball_shoot:SSLGoToBallShootEnv',
         kwargs={'field_type': 2, 'random_init': True,
                 'enter_goal_area': False},
         max_episode_steps=2400
         )

register(id='SSLStaticDefenders-v0',
         entry_point='rsoccer_gym.ssl.ssl_hw_challenge.static_defenders:SSLHWStaticDefendersEnv',
         kwargs={'field_type': 2},
         max_episode_steps=1000
         )

register(id='SSLDribbling-v0',
         entry_point='rsoccer_gym.ssl.ssl_hw_challenge.dribbling:SSLHWDribblingEnv',
         max_episode_steps=4800
         )

register(id='SSLContestedPossession-v0',
         entry_point='rsoccer_gym.ssl.ssl_hw_challenge.contested_possession:SSLContestedPossessionEnv',
         max_episode_steps=1200
         )

register(id='SSLPassEndurance-v0',
         entry_point='rsoccer_gym.ssl.ssl_hw_challenge:SSLPassEnduranceEnv',
         max_episode_steps=120
         )

register(id='SSLPassEnduranceMA-v0',
         entry_point='rsoccer_gym.ssl.ssl_hw_challenge:SSLPassEnduranceMAEnv',
         max_episode_steps=1200
         )



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: packet.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import rsoccer_gym.Simulators.pb_fira.command_fira_pb2 as command__fira__pb2
import rsoccer_gym.Simulators.pb_fira.replacement_pb2 as replacement__pb2
import rsoccer_gym.Simulators.pb_fira.common_pb2 as common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='packet.proto',
  package='fira_message.sim_to_ref',
  syntax='proto3',
  serialized_pb=_b('\n\x0cpacket.proto\x12\x17\x66ira_message.sim_to_ref\x1a\x12\x63ommand_fira.proto\x1a\x11replacement.proto\x1a\x0c\x63ommon.proto\"o\n\x06Packet\x12.\n\x03\x63md\x18\x01 \x01(\x0b\x32!.fira_message.sim_to_ref.Commands\x12\x35\n\x07replace\x18\x02 \x01(\x0b\x32$.fira_message.sim_to_ref.Replacement\"\x8d\x01\n\x0b\x45nvironment\x12\x0c\n\x04step\x18\x01 \x01(\r\x12\"\n\x05\x66rame\x18\x02 \x01(\x0b\x32\x13.fira_message.Frame\x12\"\n\x05\x66ield\x18\x03 \x01(\x0b\x32\x13.fira_message.Field\x12\x12\n\ngoals_blue\x18\x04 \x01(\r\x12\x14\n\x0cgoals_yellow\x18\x05 \x01(\r2_\n\x08Simulate\x12S\n\x08Simulate\x12\x1f.fira_message.sim_to_ref.Packet\x1a$.fira_message.sim_to_ref.Environment\"\x00\x62\x06proto3')
  ,
  dependencies=[command__fira__pb2.DESCRIPTOR,replacement__pb2.DESCRIPTOR,common__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_PACKET = _descriptor.Descriptor(
  name='Packet',
  full_name='fira_message.sim_to_ref.Packet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='cmd', full_name='fira_message.sim_to_ref.Packet.cmd', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='replace', full_name='fira_message.sim_to_ref.Packet.replace', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=94,
  serialized_end=205,
)


_ENVIRONMENT = _descriptor.Descriptor(
  name='Environment',
  full_name='fira_message.sim_to_ref.Environment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='step', full_name='fira_message.sim_to_ref.Environment.step', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='frame', full_name='fira_message.sim_to_ref.Environment.frame', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='field', full_name='fira_message.sim_to_ref.Environment.field', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='goals_blue', full_name='fira_message.sim_to_ref.Environment.goals_blue', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='goals_yellow', full_name='fira_message.sim_to_ref.Environment.goals_yellow', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=208,
  serialized_end=349,
)

_PACKET.fields_by_name['cmd'].message_type = command__fira__pb2._COMMANDS
_PACKET.fields_by_name['replace'].message_type = replacement__pb2._REPLACEMENT
_ENVIRONMENT.fields_by_name['frame'].message_type = common__pb2._FRAME
_ENVIRONMENT.fields_by_name['field'].message_type = common__pb2._FIELD
DESCRIPTOR.message_types_by_name['Packet'] = _PACKET
DESCRIPTOR.message_types_by_name['Environment'] = _ENVIRONMENT

Packet = _reflection.GeneratedProtocolMessageType('Packet', (_message.Message,), dict(
  DESCRIPTOR = _PACKET,
  __module__ = 'packet_pb2'
  # @@protoc_insertion_point(class_scope:fira_message.sim_to_ref.Packet)
  ))
_sym_db.RegisterMessage(Packet)

Environment = _reflection.GeneratedProtocolMessageType('Environment', (_message.Message,), dict(
  DESCRIPTOR = _ENVIRONMENT,
  __module__ = 'packet_pb2'
  # @@protoc_insertion_point(class_scope:fira_message.sim_to_ref.Environment)
  ))
_sym_db.RegisterMessage(Environment)


# @@protoc_insertion_point(module_scope)



import math
import random
from typing import Dict

import gym
import numpy as np
from rsoccer_gym.Entities import Frame, Robot, Ball
from rsoccer_gym.ssl.ssl_gym_base import SSLBaseEnv


class SSLContestedPossessionEnv(SSLBaseEnv):
    """The SSL robot needs to make a goal with contested possession


        Description:
            The episode starts with the enemy robot with ball possession 
            blocking the goal, the controlled robot needs to score without
            breaking division B rules
        Observation:
            Type: Box(4 + 7*n_robots_blue + 2*n_robots_yellow)
            Normalized Bounds to [-1.2, 1.2]
            Num      Observation normalized  
            0->3     Ball [X, Y, V_x, V_y]
            4->10    id 0 Blue [X, Y, sin(theta), cos(theta), v_x, v_y, v_theta]
            +2*i     id i Yellow Robot [X, Y]
        Actions:
            Type: Box(5, )
            Num     Action
            0       id 0 Blue Global X Direction Speed  (%)
            1       id 0 Blue Global Y Direction Speed  (%)
            2       id 0 Blue Angular Speed  (%)
            3       id 0 Blue Kick x Speed  (%)
            4       id 0 Blue Dribbler  (%) (true if % is positive)
            
        Reward:
            1 if goal
        Starting State:
            Enemy robot with ball possession facing away from goal
        Episode Termination:
            Goal, 30 seconds (1200 steps), or rule infraction
    """
    def __init__(self):
        super().__init__(field_type=2, n_robots_blue=1, 
                         n_robots_yellow=1, time_step=0.025)
        self.action_space = gym.spaces.Box(low=-1, high=1,
                                           shape=(5, ), dtype=np.float32)
        
        n_obs = 4 + 8*self.n_robots_blue + 2*self.n_robots_yellow
        self.observation_space = gym.spaces.Box(low=-self.NORM_BOUNDS,
                                                high=self.NORM_BOUNDS,
                                                shape=(n_obs, ),
                                                dtype=np.float32)

        # scale max dist rw to 1 Considering that max possible move rw if ball and robot are in opposite corners of field
        self.ball_dist_scale = np.linalg.norm([self.field.width, self.field.length/2])
        self.ball_grad_scale = np.linalg.norm([self.field.width/2, self.field.length/2])/4
        
        # scale max energy rw to 1 Considering that max possible energy if max robot wheel speed sent every step
        wheel_max_rad_s = 160
        max_steps = 1200
        self.energy_scale = ((wheel_max_rad_s * 4) * max_steps)

        # Limit robot speeds
        self.max_v = 2.5
        self.max_w = 10
        self.kick_speed_x = 5.0

        print('Environment initialized')

    def reset(self):
        self.reward_shaping_total = None
        return super().reset()

    def step(self, action):
        observation, reward, done, _ = super().step(action)
        return observation, reward, done, self.reward_shaping_total

    def _frame_to_observations(self):

        observation = []

        observation.append(self.norm_pos(self.frame.ball.x))
        observation.append(self.norm_pos(self.frame.ball.y))
        observation.append(self.norm_v(self.frame.ball.v_x))
        observation.append(self.norm_v(self.frame.ball.v_y))

        for i in range(self.n_robots_blue):
            observation.append(self.norm_pos(self.frame.robots_blue[i].x))
            observation.append(self.norm_pos(self.frame.robots_blue[i].y))
            observation.append(
                np.sin(np.deg2rad(self.frame.robots_blue[i].theta))
            )
            observation.append(
                np.cos(np.deg2rad(self.frame.robots_blue[i].theta))
            )
            observation.append(self.norm_v(self.frame.robots_blue[i].v_x))
            observation.append(self.norm_v(self.frame.robots_blue[i].v_y))
            observation.append(self.norm_w(self.frame.robots_blue[i].v_theta))
            observation.append(1 if self.frame.robots_blue[i].infrared else 0)

        for i in range(self.n_robots_yellow):
            observation.append(self.norm_pos(self.frame.robots_yellow[i].x))
            observation.append(self.norm_pos(self.frame.robots_yellow[i].y))

        return np.array(observation, dtype=np.float32)

    def _get_commands(self, actions):
        commands = []

        angle = self.frame.robots_blue[0].theta
        v_x, v_y, v_theta = self.convert_actions(actions, np.deg2rad(angle))
        cmd = Robot(yellow=False, id=0, v_x=v_x, v_y=v_y, v_theta=v_theta,
                    kick_v_x=self.kick_speed_x if actions[3] > 0 else 0., 
                    dribbler=True if actions[4] > 0 else False)
        commands.append(cmd)

        return commands

    def convert_actions(self, action, angle):
        """Denormalize, clip to absolute max and convert to local"""
        # Denormalize
        v_x = action[0] * self.max_v
        v_y = action[1] * self.max_v
        v_theta = action[2] * self.max_w
        # Convert to local
        v_x, v_y = v_x*np.cos(angle) + v_y*np.sin(angle),\
            -v_x*np.sin(angle) + v_y*np.cos(angle)

        # clip by max absolute
        v_norm = np.linalg.norm([v_x,v_y])
        c = v_norm < self.max_v or self.max_v / v_norm
        v_x, v_y = v_x*c, v_y*c
        
        return v_x, v_y, v_theta

    def _calculate_reward_and_done(self):
        if self.reward_shaping_total is None:
            self.reward_shaping_total = {
                'goal': 0,
                'rbt_in_gk_area': 0,
                'done_ball_out': 0,
                'done_ball_out_right': 0,
                'done_rbt_out': 0,
                'ball_dist': 0,
                'ball_grad': 0,
                'energy': 0,
                'collision': 0
            }
        reward = 0
        done = False
        
        # Field parameters
        half_len = self.field.length / 2
        half_wid = self.field.width / 2
        pen_len = self.field.penalty_length
        half_pen_wid = self.field.penalty_width / 2
        half_goal_wid = self.field.goal_width / 2
        
        ball = self.frame.ball
        robot = self.frame.robots_blue[0]
        
        def robot_in_gk_area(rbt):
            return rbt.x > half_len - pen_len and abs(rbt.y) < half_pen_wid
        
        # End episode in case of collision
        for rbt in self.frame.robots_yellow.values():
            if abs(rbt.v_x) > 0.1 or abs(rbt.v_y) > 0.1:
                self.reward_shaping_total['collision'] += 1
                done = True
        
        # Check if robot exited field right side limits
        if robot.x < -0.2 or abs(robot.y) > half_wid:
            done = True
            self.reward_shaping_total['done_rbt_out'] += 1
        # If flag is set, end episode if robot enter gk area
        elif robot_in_gk_area(robot):
            done = True
            self.reward_shaping_total['rbt_in_gk_area'] += 1
        # Check ball for ending conditions
        elif ball.x < 0 or abs(ball.y) > half_wid:
            done = True
            self.reward_shaping_total['done_ball_out'] += 1
        elif ball.x > half_len:
            done = True
            if abs(ball.y) < half_goal_wid:
                reward = 5 
                self.reward_shaping_total['goal'] += 1
            else:
                reward = 0
                self.reward_shaping_total['done_ball_out_right'] += 1
        elif self.last_frame is not None:
            ball_dist_rw = self.__ball_dist_rw() / self.ball_dist_scale
            self.reward_shaping_total['ball_dist'] += ball_dist_rw
            
            ball_grad_rw = self.__ball_grad_rw() / self.ball_grad_scale
            self.reward_shaping_total['ball_grad'] += ball_grad_rw
            
            energy_rw = -self.__energy_pen() / self.energy_scale
            self.reward_shaping_total['energy'] += energy_rw
            
            reward = reward\
                    + ball_dist_rw\
                    + ball_grad_rw\
                    + energy_rw

        done = done

        return reward, done

    def _get_initial_positions_frame(self):
        '''Returns the position of each robot and ball for the initial frame'''
        half_len = self.field.length / 2
        half_wid = self.field.width / 2
        pen_len = self.field.penalty_length
        half_pen_wid = self.field.penalty_width / 2
        pos_frame: Frame = Frame()
        def x(): return random.uniform(pen_len, half_len - pen_len)
        def y(): return random.uniform(-half_pen_wid, half_pen_wid)
        
        pos_frame.robots_blue[0] = Robot(x=0, y=0, theta=0.)
        enemy_x = x()
        enemy_y = y()

        pos_frame.ball = Ball(x=enemy_x-0.1, y=enemy_y)
        pos_frame.robots_yellow[0] = Robot(x=enemy_x, y=enemy_y, theta=180.)

        return pos_frame

    def __ball_dist_rw(self):
        assert(self.last_frame is not None)
        
        # Calculate previous ball dist
        last_ball = self.last_frame.ball
        last_robot = self.last_frame.robots_blue[0]
        last_ball_pos = np.array([last_ball.x, last_ball.y])
        last_robot_pos = np.array([last_robot.x, last_robot.y])
        last_ball_dist = np.linalg.norm(last_robot_pos - last_ball_pos)
        
        # Calculate new ball dist
        ball = self.frame.ball
        robot = self.frame.robots_blue[0]
        ball_pos = np.array([ball.x, ball.y])
        robot_pos = np.array([robot.x, robot.y])
        ball_dist = np.linalg.norm(robot_pos - ball_pos)
        
        ball_dist_rw = last_ball_dist - ball_dist
        
        if ball_dist_rw > 1:
            print("ball_dist -> ", ball_dist_rw)
            print(self.frame.ball)
            print(self.frame.robots_blue)
            print(self.frame.robots_yellow)
            print("===============================")
        
        return np.clip(ball_dist_rw, -1, 1)

    def __ball_grad_rw(self):
        assert(self.last_frame is not None)
        
        # Goal pos
        goal = np.array([self.field.length/2, 0.])
        
        # Calculate previous ball dist
        last_ball = self.last_frame.ball
        ball = self.frame.ball
        last_ball_pos = np.array([last_ball.x, last_ball.y])
        last_ball_dist = np.linalg.norm(goal - last_ball_pos)
        
        # Calculate new ball dist
        ball_pos = np.array([ball.x, ball.y])
        ball_dist = np.linalg.norm(goal - ball_pos)
        
        ball_dist_rw = last_ball_dist - ball_dist
        
        if ball_dist_rw > 1:
            print("ball_dist -> ", ball_dist_rw)
            print(self.frame.ball)
            print(self.frame.robots_blue)
            print(self.frame.robots_yellow)
            print("===============================")
        
        return np.clip(ball_dist_rw, -1, 1)

    def __energy_pen(self):
        robot = self.frame.robots_blue[0]
        
        # Sum of abs each wheel speed sent
        energy = abs(robot.v_wheel0)\
            + abs(robot.v_wheel1)\
            + abs(robot.v_wheel2)\
            + abs(robot.v_wheel3)
            
        return energy



import math
import os
import random
from typing import Dict

import gym
import numpy as np
import torch
from rsoccer_gym.Entities import Frame, Robot, Ball
from rsoccer_gym.Utils.Utils import OrnsteinUhlenbeckAction
from rsoccer_gym.vss.env_ma.opponent.model import DDPGActor
from rsoccer_gym.vss.vss_gym_base import VSSBaseEnv
from rsoccer_gym.Utils import KDTree


class VSSMAEnv(VSSBaseEnv):
    """This environment controls N robots in a VSS soccer League 3v3 match 


        Description:
        Observation:
            Type: Box(40)
            Normalized Bounds to [-1.25, 1.25]
            Num             Observation normalized  
            0               Ball X
            1               Ball Y
            2               Ball Vx
            3               Ball Vy
            4 + (7 * i)     id i Blue Robot X
            5 + (7 * i)     id i Blue Robot Y
            6 + (7 * i)     id i Blue Robot sin(theta)
            7 + (7 * i)     id i Blue Robot cos(theta)
            8 + (7 * i)     id i Blue Robot Vx
            9  + (7 * i)    id i Blue Robot Vy
            10 + (7 * i)    id i Blue Robot v_theta
            25 + (5 * i)    id i Yellow Robot X
            26 + (5 * i)    id i Yellow Robot Y
            27 + (5 * i)    id i Yellow Robot Vx
            28 + (5 * i)    id i Yellow Robot Vy
            29 + (5 * i)    id i Yellow Robot v_theta
        Actions:
            Type: Box(N, 2)
            For each blue robot in control:
                Num     Action
                0       Left Wheel Speed  (%)
                1       Right Wheel Speed (%)
        Reward:
            Sum of Rewards:
                For all robots:
                    Goal
                    Ball Potential Gradient
                Individual:
                    Move to Ball
                    Energy Penalty
        Starting State:
            Randomized Robots and Ball initial Position
        Episode Termination:
            5 minutes match time
    """

    def __init__(self, n_robots_control=3):
        super().__init__(field_type=0, n_robots_blue=3, n_robots_yellow=3,
                         time_step=0.025)

        self.n_robots_control = n_robots_control
        self.action_space = gym.spaces.Box(low=-1,
                                           high=1,
                                           shape=(n_robots_control, 2))
        self.observation_space = gym.spaces.Box(low=-self.NORM_BOUNDS,
                                                high=self.NORM_BOUNDS,
                                                shape=(n_robots_control, 40),
                                                dtype=np.float32)

        # Initialize Class Atributes
        self.previous_ball_potential = None
        self.actions: Dict = None
        self.reward_shaping_total = None
        self.v_wheel_deadzone = 0.05
        
        self.ou_actions = []
        for i in range(self.n_robots_blue + self.n_robots_yellow):
            self.ou_actions.append(
                OrnsteinUhlenbeckAction(self.action_space, dt=self.time_step)
            )

        print('Environment initialized')

    def reset(self):
        self.actions = None
        self.reward_shaping_total = None
        self.previous_ball_potential = None
        for ou in self.ou_actions:
            ou.reset()

        return super().reset()

    def step(self, action):
        observation, reward, done, _ = super().step(action)
        return observation, reward, done, self.reward_shaping_total

    def get_rotated_obs(self):
        robots_dict = dict()
        for i in range(self.n_robots_blue):
            robots_dict[i] = list()
            robots_dict[i].append(self.norm_pos(self.frame.robots_blue[i].x))
            robots_dict[i].append(self.norm_pos(self.frame.robots_blue[i].y))
            robots_dict[i].append(
                np.sin(np.deg2rad(self.frame.robots_blue[i].theta))
            )
            robots_dict[i].append(
                np.cos(np.deg2rad(self.frame.robots_blue[i].theta))
            )
            robots_dict[i].append(self.norm_v(self.frame.robots_blue[i].v_x))
            robots_dict[i].append(self.norm_v(self.frame.robots_blue[i].v_y))
            robots_dict[i].append(self.norm_w(self.frame.robots_blue[i].v_theta))

        rotaded_obs = list()
        for i in range(self.n_robots_control):
            aux_dict = {}
            aux_dict.update(robots_dict)
            rotated = list()
            rotated = rotated + aux_dict.pop(i)
            teammates = list(aux_dict.values())
            for teammate in teammates:
                rotated = rotated + teammate
            rotaded_obs.append(rotated)

        return rotaded_obs

    def _frame_to_observations(self):

        observations = list()
        robots = self.get_rotated_obs()
        for idx in range(self.n_robots_control):
            observation = []

            observation.append(self.norm_pos(self.frame.ball.x))
            observation.append(self.norm_pos(self.frame.ball.y))
            observation.append(self.norm_v(self.frame.ball.v_x))
            observation.append(self.norm_v(self.frame.ball.v_y))

            observation += robots[idx]

            for i in range(self.n_robots_yellow):
                observation.append(self.norm_pos(self.frame.robots_yellow[i].x))
                observation.append(self.norm_pos(self.frame.robots_yellow[i].y))
                observation.append(self.norm_v(self.frame.robots_yellow[i].v_x))
                observation.append(self.norm_v(self.frame.robots_yellow[i].v_y))
                observation.append(self.norm_w(self.frame.robots_yellow[i].v_theta))

            observations.append(np.array(observation, dtype=np.float32))

        observations = np.array(observations)
        return observations

    def _get_commands(self, actions):
        commands = []
        self.actions = {}

        # Send random commands to the other robots
        for i in range(self.n_robots_control):
            self.actions[i] = actions[i]
            v_wheel0, v_wheel1 = self._actions_to_v_wheels(actions[i])
            commands.append(Robot(yellow=False, id=i, v_wheel0=v_wheel0,
                                  v_wheel1=v_wheel1))

        for i in range(self.n_robots_control, self.n_robots_blue):
            actions = self.ou_actions[i].sample()
            v_wheel0, v_wheel1 = self._actions_to_v_wheels(actions[0])
            commands.append(Robot(yellow=False, id=i, v_wheel0=v_wheel0,
                                  v_wheel1=v_wheel1))

        for i in range(self.n_robots_yellow):
            actions = self.ou_actions[self.n_robots_blue+i].sample()
            v_wheel0, v_wheel1 = self._actions_to_v_wheels(actions[0])
            commands.append(Robot(yellow=True, id=i, v_wheel0=v_wheel0,
                                  v_wheel1=v_wheel1))

        return commands

    def _calculate_reward_and_done(self):
        reward = {f'robot_{i}': 0 for i in range(self.n_robots_control)}
        goal = False
        w_move = 0.2
        w_ball_grad = 0.8
        w_energy = 2e-4
        if self.reward_shaping_total is None:
            self.reward_shaping_total = {'goal_score': 0, 'ball_grad': 0,
                                         'goals_blue': 0, 'goals_yellow': 0}
            for i in range(self.n_robots_control):
                self.reward_shaping_total[f'robot_{i}'] = {
                    'move': 0, 'energy': 0}

        # Check if goal ocurred
        if self.frame.ball.x > (self.field.length / 2):
            self.reward_shaping_total['goal_score'] += 1
            self.reward_shaping_total['goals_blue'] += 1
            for i in range(self.n_robots_control):
                reward[f'robot_{i}'] = 10
            goal = True
        elif self.frame.ball.x < -(self.field.length / 2):
            self.reward_shaping_total['goal_score'] -= 1
            self.reward_shaping_total['goals_yellow'] += 1
            for i in range(self.n_robots_control):
                reward[f'robot_{i}'] = -10
            goal = True
        else:

            if self.last_frame is not None:
                # Calculate ball potential
                grad_ball_potential = self._ball_grad()
                self.reward_shaping_total['ball_grad'] += w_ball_grad * grad_ball_potential  # noqa
                for idx in range(self.n_robots_control):
                    # Calculate Move ball
                    move_reward = self._move_reward(robot_idx=idx)
                    # Calculate Energy penalty
                    energy_penalty = self._energy_penalty(robot_idx=idx)

                    rew = w_ball_grad * grad_ball_potential + \
                        w_move * move_reward + \
                        w_energy * energy_penalty

                    reward[f'robot_{idx}'] += rew
                    self.reward_shaping_total[f'robot_{idx}']['move'] += w_move * move_reward  # noqa
                    self.reward_shaping_total[f'robot_{idx}']['energy'] += w_energy * energy_penalty  # noqa

        return reward, goal

    def _get_initial_positions_frame(self):
        '''Returns the position of each robot and ball for the initial frame'''
        field_half_length = self.field.length / 2
        field_half_width = self.field.width / 2

        def x(): return random.uniform(-field_half_length + 0.1,
                                       field_half_length - 0.1)

        def y(): return random.uniform(-field_half_width + 0.1,
                                       field_half_width - 0.1)

        def theta(): return random.uniform(0, 360)

        pos_frame: Frame = Frame()

        pos_frame.ball = Ball(x=x(), y=y())

        min_dist = 0.1

        places = KDTree()
        places.insert((pos_frame.ball.x, pos_frame.ball.y))
        
        for i in range(self.n_robots_blue):
            pos = (x(), y())
            while places.get_nearest(pos)[1] < min_dist:
                pos = (x(), y())

            places.insert(pos)
            pos_frame.robots_blue[i] = Robot(x=pos[0], y=pos[1], theta=theta())

        for i in range(self.n_robots_yellow):
            pos = (x(), y())
            while places.get_nearest(pos)[1] < min_dist:
                pos = (x(), y())

            places.insert(pos)
            pos_frame.robots_yellow[i] = Robot(x=pos[0], y=pos[1], theta=theta())

        return pos_frame

    def _actions_to_v_wheels(self, actions):
        left_wheel_speed = actions[0] * self.max_v
        right_wheel_speed = actions[1] * self.max_v

        left_wheel_speed, right_wheel_speed = np.clip(
            (left_wheel_speed, right_wheel_speed), -self.max_v, self.max_v
        )

        # Deadzone
        if -self.v_wheel_deadzone < left_wheel_speed < self.v_wheel_deadzone:
            left_wheel_speed = 0

        if -self.v_wheel_deadzone < right_wheel_speed < self.v_wheel_deadzone:
            right_wheel_speed = 0

        # Convert to rad/s
        left_wheel_speed /= self.field.rbt_wheel_radius
        right_wheel_speed /= self.field.rbt_wheel_radius

        return left_wheel_speed , right_wheel_speed

    def _ball_grad(self):
        '''Calculate ball potential gradient
        Difference of potential of the ball in time_step seconds.
        '''
        # Calculate ball potential
        length_cm = self.field.length * 100
        half_lenght = (self.field.length / 2.0)\
            + self.field.goal_depth

        # distance to defence
        dx_d = (half_lenght + self.frame.ball.x) * 100
        # distance to attack
        dx_a = (half_lenght - self.frame.ball.x) * 100
        dy = (self.frame.ball.y) * 100

        dist_1 = -math.sqrt(dx_a ** 2 + 2 * dy ** 2)
        dist_2 = math.sqrt(dx_d ** 2 + 2 * dy ** 2)
        ball_potential = ((dist_1 + dist_2) / length_cm - 1) / 2

        grad_ball_potential = 0
        # Calculate ball potential gradient
        # = actual_potential - previous_potential
        if self.previous_ball_potential is not None:
            diff = ball_potential - self.previous_ball_potential
            grad_ball_potential = np.clip(diff * 3 / self.time_step,
                                          -5.0, 5.0)

        self.previous_ball_potential = ball_potential

        return grad_ball_potential

    def _move_reward(self, robot_idx: int):
        '''Calculate Move to ball reward

        Cosine between the robot vel vector and the vector robot -> ball.
        This indicates rather the robot is moving towards the ball or not.
        '''

        ball = np.array([self.frame.ball.x, self.frame.ball.y])
        robot = np.array([self.frame.robots_blue[robot_idx].x,
                          self.frame.robots_blue[robot_idx].y])
        robot_vel = np.array([self.frame.robots_blue[robot_idx].v_x,
                              self.frame.robots_blue[robot_idx].v_y])
        robot_ball = ball - robot
        robot_ball = robot_ball/np.linalg.norm(robot_ball)

        move_reward = np.dot(robot_ball, robot_vel)

        move_reward = np.clip(move_reward / 0.4, -5.0, 5.0)
        return move_reward

    def _energy_penalty(self, robot_idx: int):
        '''Calculates the energy penalty'''

        en_penalty_1 = abs(self.sent_commands[robot_idx].v_wheel0)
        en_penalty_2 = abs(self.sent_commands[robot_idx].v_wheel1)
        energy_penalty = - (en_penalty_1 + en_penalty_2)
        return energy_penalty


class VSSMAOpp(VSSMAEnv):

    def __init__(self, n_robots_control=3):
        super().__init__(n_robots_control=n_robots_control)
        self.load_opp()

    def load_opp(self):
        device = torch.device('cpu')
        atk_path = os.path.dirname(os.path.realpath(__file__))\
            + '/opponent/opp.pth'
        self.opp = DDPGActor(40, 2)
        atk_checkpoint = torch.load(atk_path, map_location=device)
        self.opp.load_state_dict(atk_checkpoint['state_dict_act'])
        self.opp.eval()

    def _opp_obs(self):
        observation = []
        observation.append(self.norm_pos(-self.frame.ball.x))
        observation.append(self.norm_pos(self.frame.ball.y))
        observation.append(self.norm_v(-self.frame.ball.v_x))
        observation.append(self.norm_v(self.frame.ball.v_y))

        #  we reflect the side that the opp is attacking,
        #  so that he will attack towards the goal where the goalkeeper is
        for i in range(self.n_robots_yellow):
            observation.append(self.norm_pos(-self.frame.robots_yellow[i].x))
            observation.append(self.norm_pos(self.frame.robots_yellow[i].y))

            observation.append(
                np.sin(np.deg2rad(self.frame.robots_yellow[i].theta))
            )
            observation.append(
                -np.cos(np.deg2rad(self.frame.robots_yellow[i].theta))
            )
            observation.append(self.norm_v(-self.frame.robots_yellow[i].v_x))
            observation.append(self.norm_v(self.frame.robots_yellow[i].v_y))

            observation.append(self.norm_w(-self.frame.robots_yellow[i].v_theta))

        for i in range(self.n_robots_blue):
            observation.append(self.norm_pos(-self.frame.robots_blue[i].x))
            observation.append(self.norm_pos(self.frame.robots_blue[i].y))
            observation.append(self.norm_v(-self.frame.robots_blue[i].v_x))
            observation.append(self.norm_v(self.frame.robots_blue[i].v_y))
            observation.append(self.norm_w(-self.frame.robots_blue[i].v_theta))

        return np.array(observation, dtype=np.float32)

    def _get_commands(self, actions):
        commands = []
        self.actions = {}

        for i in range(self.n_robots_control):
            self.actions[i] = actions[i]
            v_wheel0, v_wheel1 = self._actions_to_v_wheels(actions[i])
            commands.append(Robot(yellow=False, id=i, v_wheel0=v_wheel0,
                                  v_wheel1=v_wheel1))

        for i in range(self.n_robots_control, self.n_robots_blue):
            actions = self.action_space.sample()
            v_wheel0, v_wheel1 = self._actions_to_v_wheels(actions)
            commands.append(Robot(yellow=False, id=i, v_wheel0=v_wheel0,
                                  v_wheel1=v_wheel1))

        atk_action = self.opp.get_action(self._opp_obs())
        v_wheel0, v_wheel1 = self._actions_to_v_wheels(atk_action)
        commands.append(Robot(yellow=True, id=0, v_wheel0=v_wheel1,
                              v_wheel1=v_wheel0))
        for i in range(1, self.n_robots_yellow):
            actions = self.action_space.sample()
            v_wheel0, v_wheel1 = self._actions_to_v_wheels(actions)
            commands.append(Robot(yellow=True, id=i, v_wheel0=v_wheel0,
                                  v_wheel1=v_wheel1))

        return commands



import numpy as np
import gym

# Base on baselines implementation
class OrnsteinUhlenbeckAction(object):
    def __init__(self, action_space, theta=.17, dt=0.025, x0=None):
        self.theta = theta
        self.mu = (action_space.high + action_space.low) / 2
        self.sigma = (action_space.high - self.mu) / 2
        self.dt = dt
        self.x0 = x0
        self.reset()

    def sample(self):
        x = self.x_prev + self.theta * (self.mu - self.x_prev) * self.dt + self.sigma * np.sqrt(self.dt) * np.random.normal(size=self.mu.shape)
        self.x_prev = x
        return x

    def reset(self):
        self.x_prev = self.x0 if self.x0 is not None else np.zeros_like(self.mu)

    def __repr__(self):
        return 'OrnsteinUhlenbeckActionNoise(mu={}, sigma={})'.format(self.mu, self.sigma)



import math


def closest_node(values, node1, node2):

    if node1 is None:
        return node2, node2.distance2_to(values) if node2 is not None else math.inf

    if node2 is None:
        return node1, node1.distance2_to(values) if node1 is not None else math.inf

    node1_dist2 = node1.distance2_to(values)
    node2_dist2 = node2.distance2_to(values)

    if node1_dist2 < node2_dist2:
        return node1, node1_dist2
    else:
        return node2, node2_dist2


class KDTree:
    class KDTreeNode:

        def __init__(self, values, left=None, right=None):
            self.values = values
            self.left = left
            self.right = right

        def insert(self, values, depth=0):
            if self.values is None:
                self.values = values
            else:
                if values[depth % len(values)] < self.values[depth % len(self.values)]:
                    if self.left is None:
                        self.left = KDTree.KDTreeNode(values)
                    else:
                        self.left.insert(values, depth+1)
                else:
                    if self.right is None:
                        self.right = KDTree.KDTreeNode(values)
                    else:
                        self.right.insert(values, depth+1)

        def distance2_to(self, values):
            d2 = 0
            for i in range(len(values)):
                d2 += (values[i] - self.values[i])**2

            return d2

        def get_nearest(self, values, depth=0):
            if self.values is None:
                return None, math.inf

            if self.left is None and self.right is None:
                return self, self.distance2_to(values)

            if values[depth % len(values)] < self.values[depth % len(self.values)]:
                next_branch = self.left
                other_branch = self.right
            else:
                next_branch = self.left
                other_branch = self.right

            if next_branch is not None:
                other, _ = next_branch.get_nearest(values, depth+1)
                closest, closest_dist2 = closest_node(values, other, self)
            else:
                closest, closest_dist2 = self, self.distance2_to(values)

            line_dist = values[depth % len(values)] - self.values[depth % len(self.values)]

            if other_branch is not None:
                if closest_dist2 >= line_dist**2:
                    other, _ = other_branch.get_nearest(values, depth+1)
                    closest, closest_dist2 = closest_node(values, other, closest)

            return closest, closest_dist2

    def __init__(self):
        self.root = KDTree.KDTreeNode(None)

    def insert(self, values):
        self.root.insert(values)

    def get_nearest(self, values):
        node, dist2 = self.root.get_nearest(values)
        return node.values, math.sqrt(dist2)



from dataclasses import dataclass

@dataclass()
class Field:
    length: float
    width: float
    penalty_length: float
    penalty_width: float
    goal_width: float
    goal_depth: float
    ball_radius: float
    rbt_distance_center_kicker: float
    rbt_kicker_thickness: float
    rbt_kicker_width: float
    rbt_wheel0_angle: float
    rbt_wheel1_angle: float
    rbt_wheel2_angle: float
    rbt_wheel3_angle: float
    rbt_radius: float
    rbt_wheel_radius: float
    rbt_motor_max_rpm: float


# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: replacement.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


import rsoccer_gym.Simulators.pb_fira.common_pb2 as common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='replacement.proto',
  package='fira_message.sim_to_ref',
  syntax='proto3',
  serialized_pb=_b('\n\x11replacement.proto\x12\x17\x66ira_message.sim_to_ref\x1a\x0c\x63ommon.proto\"]\n\x10RobotReplacement\x12%\n\x08position\x18\x01 \x01(\x0b\x32\x13.fira_message.Robot\x12\x12\n\nyellowteam\x18\x05 \x01(\x08\x12\x0e\n\x06turnon\x18\x06 \x01(\x08\"?\n\x0f\x42\x61llReplacement\x12\t\n\x01x\x18\x01 \x01(\x01\x12\t\n\x01y\x18\x02 \x01(\x01\x12\n\n\x02vx\x18\x03 \x01(\x01\x12\n\n\x02vy\x18\x04 \x01(\x01\"\x80\x01\n\x0bReplacement\x12\x36\n\x04\x62\x61ll\x18\x01 \x01(\x0b\x32(.fira_message.sim_to_ref.BallReplacement\x12\x39\n\x06robots\x18\x02 \x03(\x0b\x32).fira_message.sim_to_ref.RobotReplacementb\x06proto3')
  ,
  dependencies=[common__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_ROBOTREPLACEMENT = _descriptor.Descriptor(
  name='RobotReplacement',
  full_name='fira_message.sim_to_ref.RobotReplacement',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='position', full_name='fira_message.sim_to_ref.RobotReplacement.position', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='yellowteam', full_name='fira_message.sim_to_ref.RobotReplacement.yellowteam', index=1,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='turnon', full_name='fira_message.sim_to_ref.RobotReplacement.turnon', index=2,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=60,
  serialized_end=153,
)


_BALLREPLACEMENT = _descriptor.Descriptor(
  name='BallReplacement',
  full_name='fira_message.sim_to_ref.BallReplacement',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='x', full_name='fira_message.sim_to_ref.BallReplacement.x', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='y', full_name='fira_message.sim_to_ref.BallReplacement.y', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='vx', full_name='fira_message.sim_to_ref.BallReplacement.vx', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='vy', full_name='fira_message.sim_to_ref.BallReplacement.vy', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=155,
  serialized_end=218,
)


_REPLACEMENT = _descriptor.Descriptor(
  name='Replacement',
  full_name='fira_message.sim_to_ref.Replacement',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ball', full_name='fira_message.sim_to_ref.Replacement.ball', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='robots', full_name='fira_message.sim_to_ref.Replacement.robots', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=221,
  serialized_end=349,
)

_ROBOTREPLACEMENT.fields_by_name['position'].message_type = common__pb2._ROBOT
_REPLACEMENT.fields_by_name['ball'].message_type = _BALLREPLACEMENT
_REPLACEMENT.fields_by_name['robots'].message_type = _ROBOTREPLACEMENT
DESCRIPTOR.message_types_by_name['RobotReplacement'] = _ROBOTREPLACEMENT
DESCRIPTOR.message_types_by_name['BallReplacement'] = _BALLREPLACEMENT
DESCRIPTOR.message_types_by_name['Replacement'] = _REPLACEMENT

RobotReplacement = _reflection.GeneratedProtocolMessageType('RobotReplacement', (_message.Message,), dict(
  DESCRIPTOR = _ROBOTREPLACEMENT,
  __module__ = 'replacement_pb2'
  # @@protoc_insertion_point(class_scope:fira_message.sim_to_ref.RobotReplacement)
  ))
_sym_db.RegisterMessage(RobotReplacement)

BallReplacement = _reflection.GeneratedProtocolMessageType('BallReplacement', (_message.Message,), dict(
  DESCRIPTOR = _BALLREPLACEMENT,
  __module__ = 'replacement_pb2'
  # @@protoc_insertion_point(class_scope:fira_message.sim_to_ref.BallReplacement)
  ))
_sym_db.RegisterMessage(BallReplacement)

Replacement = _reflection.GeneratedProtocolMessageType('Replacement', (_message.Message,), dict(
  DESCRIPTOR = _REPLACEMENT,
  __module__ = 'replacement_pb2'
  # @@protoc_insertion_point(class_scope:fira_message.sim_to_ref.Replacement)
  ))
_sym_db.RegisterMessage(Replacement)


# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: command.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='command.proto',
  package='fira_message.sim_to_ref',
  syntax='proto3',
  serialized_pb=_b('\n\rcommand.proto\x12\x17\x66ira_message.sim_to_ref\"R\n\x07\x43ommand\x12\n\n\x02id\x18\x01 \x01(\r\x12\x12\n\nyellowteam\x18\x02 \x01(\x08\x12\x12\n\nwheel_left\x18\x06 \x01(\x01\x12\x13\n\x0bwheel_right\x18\x07 \x01(\x01\"D\n\x08\x43ommands\x12\x38\n\x0erobot_commands\x18\x01 \x03(\x0b\x32 .fira_message.sim_to_ref.Commandb\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_COMMAND = _descriptor.Descriptor(
  name='Command',
  full_name='fira_message.sim_to_ref.Command',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='fira_message.sim_to_ref.Command.id', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='yellowteam', full_name='fira_message.sim_to_ref.Command.yellowteam', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='wheel_left', full_name='fira_message.sim_to_ref.Command.wheel_left', index=2,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='wheel_right', full_name='fira_message.sim_to_ref.Command.wheel_right', index=3,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=42,
  serialized_end=124,
)


_COMMANDS = _descriptor.Descriptor(
  name='Commands',
  full_name='fira_message.sim_to_ref.Commands',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='robot_commands', full_name='fira_message.sim_to_ref.Commands.robot_commands', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=126,
  serialized_end=194,
)

_COMMANDS.fields_by_name['robot_commands'].message_type = _COMMAND
DESCRIPTOR.message_types_by_name['Command'] = _COMMAND
DESCRIPTOR.message_types_by_name['Commands'] = _COMMANDS

Command = _reflection.GeneratedProtocolMessageType('Command', (_message.Message,), dict(
  DESCRIPTOR = _COMMAND,
  __module__ = 'command_pb2'
  # @@protoc_insertion_point(class_scope:fira_message.sim_to_ref.Command)
  ))
_sym_db.RegisterMessage(Command)

Commands = _reflection.GeneratedProtocolMessageType('Commands', (_message.Message,), dict(
  DESCRIPTOR = _COMMANDS,
  __module__ = 'command_pb2'
  # @@protoc_insertion_point(class_scope:fira_message.sim_to_ref.Commands)
  ))
_sym_db.RegisterMessage(Commands)


# @@protoc_insertion_point(module_scope)



import math
import random
from typing import Dict

import gym
import numpy as np
from rsoccer_gym.Entities import Frame, Robot, Ball
from rsoccer_gym.ssl.ssl_gym_base import SSLBaseEnv


class SSLHWDribblingEnv(SSLBaseEnv):
    """The SSL robot needs navigate a course while keeping the ball


        Description:
            The robot must navigate through a field with robots as obstacles,
            while keeping the ball. The obstacles are placed on a straight line,
            the course will be in a zigzag configuration, having checkpoints
            defined as the space between two robots, the last checkpoint needs
            to be passed three times
        Observation:
            Type: Box(4 + 7*n_robots_blue + 2*n_robots_yellow)
            Normalized Bounds to [-1.2, 1.2]
            Num      Observation normalized  
            0->3     Ball [X, Y, V_x, V_y]
            4->10    id 0 Blue [X, Y, sin(theta), cos(theta), v_x, v_y, v_theta]
            +2*i     id i Yellow Robot [X, Y]
        Actions:
            Type: Box(4,)
            Num     Action
            0       id 0 Blue Global X Direction Speed  (%)
            1       id 0 Blue Global Y Direction Speed  (%)
            2       id 0 Blue Angular Speed  (%)
            3       id 0 Blue Dribbler  (%) (true if % is positive)
            
        Reward:
            1 every time the robot passes a checkpoint
        Starting State:
            Robot starts with the ball and obstacles are spaced with 
            pre defined values.
        Episode Termination:
            Course completed, 2 minutes (4800 steps), robot exits course limits or robot 
            reverse a checkpoint
    """

    def __init__(self):
        super().__init__(field_type=2, n_robots_blue=1, 
                         n_robots_yellow=4, time_step=0.025)
        self.action_space = gym.spaces.Box(low=-1, high=1,
                                           shape=(4, ), dtype=np.float32)
        
        n_obs = 5 + 8*self.n_robots_blue + 2*self.n_robots_yellow
        self.observation_space = gym.spaces.Box(low=-self.NORM_BOUNDS,
                                                high=self.NORM_BOUNDS,
                                                shape=(n_obs, ),
                                                dtype=np.float32)
        
        self.checkpoints_count = 0
        # Checkpoints nodes positions
        self.node_0 = -0.5 
        self.node_1 = -1.
        self.node_2 = -1.5
        self.node_3 = -2.
        self.field_margin = 1
        
        # Limit robot speeds
        self.max_v = 2.5
        self.max_w = 10
        
        print('Environment initialized')

    def reset(self):
        self.checkpoints_count = 0
        return super().reset()

    def _frame_to_observations(self):

        observation = []

        observation.append(((self.checkpoints_count/6)*2)-1)
        observation.append(self.norm_pos(self.frame.ball.x))
        observation.append(self.norm_pos(self.frame.ball.y))
        observation.append(self.norm_v(self.frame.ball.v_x))
        observation.append(self.norm_v(self.frame.ball.v_y))

        for i in range(self.n_robots_blue):
            observation.append(self.norm_pos(self.frame.robots_blue[i].x))
            observation.append(self.norm_pos(self.frame.robots_blue[i].y))
            observation.append(
                np.sin(np.deg2rad(self.frame.robots_blue[i].theta))
            )
            observation.append(
                np.cos(np.deg2rad(self.frame.robots_blue[i].theta))
            )
            observation.append(self.norm_v(self.frame.robots_blue[i].v_x))
            observation.append(self.norm_v(self.frame.robots_blue[i].v_y))
            observation.append(self.norm_w(self.frame.robots_blue[i].v_theta))
            observation.append(1 if self.frame.robots_blue[i].infrared else -1)

        for i in range(self.n_robots_yellow):
            observation.append(self.norm_pos(self.frame.robots_yellow[i].x))
            observation.append(self.norm_pos(self.frame.robots_yellow[i].y))

        return np.array(observation, dtype=np.float32)

    def _get_commands(self, actions):
        commands = []

        angle = self.frame.robots_blue[0].theta
        v_x, v_y, v_theta = self.convert_actions(actions, np.deg2rad(angle))
        cmd = Robot(yellow=False, id=0, v_x=v_x, v_y=v_y, v_theta=v_theta,
                    dribbler=True if actions[3] > 0 else False)
        commands.append(cmd)

        return commands

    def convert_actions(self, action, angle):
        """Denormalize, clip to absolute max and convert to local"""
        # Denormalize
        v_x = action[0] * self.max_v
        v_y = action[1] * self.max_v
        v_theta = action[2] * self.max_w
        # Convert to local
        v_x, v_y = v_x*np.cos(angle) + v_y*np.sin(angle),\
            -v_x*np.sin(angle) + v_y*np.cos(angle)

        # clip by max absolute
        v_norm = np.linalg.norm([v_x,v_y])
        c = v_norm < self.max_v or self.max_v / v_norm
        v_x, v_y = v_x*c, v_y*c
        
        return v_x, v_y, v_theta

    def _calculate_reward_and_done(self):
        reward = 0
        done = False
        
        ball = self.frame.ball
        last_ball = None or self.last_frame.ball
        robot = self.frame.robots_blue[0]
        
        # End episode in case of collision
        for rbt in self.frame.robots_yellow.values():
            if abs(rbt.v_x) > 0.05 or abs(rbt.v_y) > 0.05:
                done = True
        
        def robot_out_of_bounds(rbt):
            if rbt.x < self.node_3 - self.field_margin or rbt.x > self.field_margin:
                return True
            if abs(rbt.y) > self.field_margin:
                return True
            return False
        
        if robot_out_of_bounds(robot):
            done = True
        elif last_ball:
            if self.checkpoints_count == 0:
                if ball.x < self.node_0 and ball.x > self.node_1:
                    if last_ball.y >= 0 and ball.y < 0:
                        reward = 1
                        self.checkpoints_count += 1
            elif self.checkpoints_count == 1:
                if ball.x < self.node_1 and ball.x > self.node_2:
                    if last_ball.y < 0 and ball.y >= 0:
                        reward = 1
                        self.checkpoints_count += 1
            elif self.checkpoints_count >= 2:
                if self.checkpoints_count % 2 == 0:
                    if ball.x < self.node_2 and ball.x > self.node_3:
                        if last_ball.y >= 0 and ball.y < 0:
                            reward = 1
                            self.checkpoints_count += 1
                            if self.checkpoints_count == 7:
                                done = True
                        elif last_ball.y < 0 and ball.y >= 0:
                            done = True
                else:
                    if ball.x > self.node_3 - self.field_margin and ball.x < self.node_3:
                        if last_ball.y < 0 and ball.y >= 0:
                            reward = 1
                            self.checkpoints_count += 1

        done = done

        return reward, done
    
    def _get_initial_positions_frame(self):
        '''Returns the position of each robot and ball for the initial frame'''
        # TODO

        pos_frame: Frame = Frame()

        pos_frame.ball = Ball(x=-0.1, y=0.)

        pos_frame.robots_blue[0] = Robot(x=0., y=0., theta=180.)
        
        pos_frame.robots_yellow[0] = Robot(x=self.node_0, y=0., theta=180.)
        pos_frame.robots_yellow[1] = Robot(x=self.node_1, y=0., theta=180.)
        pos_frame.robots_yellow[2] = Robot(x=self.node_2, y=0., theta=180.)
        pos_frame.robots_yellow[3] = Robot(x=self.node_3, y=0., theta=180.)

        return pos_frame



import socket
from typing import Dict, List

import numpy as np
from rsoccer_gym.Entities import Robot
from rsoccer_gym.Entities.Frame import FramePB
from rsoccer_gym.Simulators.rsim import RSim

import rsoccer_gym.Simulators.pb_fira.packet_pb2 as packet_pb2
from rsoccer_gym.Simulators.pb_fira.state_pb2 import *


class Fira(RSim):
    def __init__(
        self,
        vision_ip="224.0.0.1",
        vision_port=10002,
        cmd_ip="127.0.0.1",
        cmd_port=20011,
    ):
        """
        Init SSLClient object.
        Extended description of function.
        Parameters
        ----------
        ip : str
            Multicast IP in format '255.255.255.255'. 
        port : int
            Port up to 1024. 
        """

        self.vision_ip = vision_ip
        self.vision_port = vision_port
        self.com_ip = cmd_ip
        self.com_port = cmd_port
        self.com_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.com_address = (self.com_ip, self.com_port)

        self.vision_sock = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP
        )
        self.vision_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.vision_sock.setsockopt(
            socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 128
        )
        self.vision_sock.setsockopt(
            socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, 1
        )
        self.vision_sock.bind((self.vision_ip, self.vision_port))
        self.linear_speed_range = 1.15
        self.robot_wheel_radius = 0.026

    def get_field_params(self):
        return {'field_width': 1.3, 'field_length': 1.5,
                'penalty_width': 0.7, 'penalty_length': 0.15,
                'goal_width': 0.4, 'goal_depth': 0.1}

    def stop(self):
        pass

    def reset(self, frame: FramePB):
        placement_pos = self._placement_dict_from_frame(frame)
        pkt = packet_pb2.Packet()

        ball_pos = placement_pos["ball_pos"][:2]
        ball_pkt = pkt.replace.ball
        ball_pkt.x = ball_pos[0]
        ball_pkt.y = ball_pos[1]

        robots_pkt = pkt.replace.robots
        for i, robot in enumerate(placement_pos["blue_robots_pos"]):
            rep_rob = robots_pkt.add()
            rep_rob.position.robot_id = i+1
            rep_rob.position.x = robot[0]
            rep_rob.position.y = robot[1]
            rep_rob.position.orientation = robot[2]
            rep_rob.yellowteam = False
            rep_rob.turnon = True

        for i, robot in enumerate(placement_pos["yellow_robots_pos"]):
            rep_rob = robots_pkt.add()
            rep_rob.position.robot_id = i+1
            rep_rob.position.x = robot[0]
            rep_rob.position.y = robot[1]
            rep_rob.position.orientation = robot[2]
            rep_rob.yellowteam = True
            rep_rob.turnon = True

        # send commands
        data = pkt.SerializeToString()
        self.com_socket.sendto(data, self.com_address)

    def get_frame(self):
        """Receive package and decode."""
        data, _ = self.vision_sock.recvfrom(1024)
        decoded_data = packet_pb2.Environment().FromString(data)
        frame = FramePB()
        frame.parse(decoded_data)
        return frame

    def send_commands(self, commands):
        # prepare commands
        pkt = packet_pb2.Packet()
        d = pkt.cmd.robot_commands

        # send wheel speed commands for each robot
        for cmd in commands:
            robot = d.add()
            robot.id = cmd.id
            robot.yellowteam = cmd.yellow

            # convert from linear speed to angular speed
            robot.wheel_left = cmd.v_wheel1 / self.robot_wheel_radius
            robot.wheel_right = cmd.v_wheel2 / self.robot_wheel_radius

        # send commands
        data = pkt.SerializeToString()
        self.com_socket.sendto(data, self.com_address)

    def _placement_dict_from_frame(self, frame: FramePB):
        replacement_pos: Dict[str, np.ndarray] = {}

        ball_pos: List[float] = [
            frame.ball.x,
            frame.ball.y,
            frame.ball.v_x,
            frame.ball.v_y,
        ]
        replacement_pos["ball_pos"] = np.array(ball_pos)

        blue_pos: List[List[float]] = []
        for robot in frame.robots_blue.values():
            robot_pos: List[float] = [robot.x, robot.y, robot.theta]
            blue_pos.append(robot_pos)
        replacement_pos["blue_robots_pos"] = np.array(blue_pos)

        yellow_pos: List[List[float]] = []
        for robot in frame.robots_yellow.values():
            robot_pos: List[float] = [robot.x, robot.y, robot.theta]
            yellow_pos.append(robot_pos)
        replacement_pos["yellow_robots_pos"] = np.array(yellow_pos)

        return replacement_pos
